# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTIBILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import bpy
from bpy.app.handlers import frame_change_pre, load_post, render_pre, save_post
from bpy.props import CollectionProperty, PointerProperty
from .. import session
from ..handlers import render_pre_handler, update_entities
from ..utils import add_current_module_to_path
from . import node_menu, ops, panel, pref, props

all_classes = (
    panel.CLASSES
    + ops.CLASSES
    + props.CLASSES
    + pref.CLASSES
    + session.CLASSES
    + node_menu.CLASSES
)

_is_registered = False


def _test_register():
    try:
        register()
    except Exception as e:
        print(e)
        unregister()
        register()


def register():
    global _is_registered

    if _is_registered:
        return

    # register all of the import operators
    for op in all_classes:
        try:
            bpy.utils.register_class(op)
        except Exception as e:
            print(e)
            pass
    add_current_module_to_path()
    bpy.types.NODE_MT_add.append(node_menu.add_node_menu)
    bpy.types.VIEW3D_MT_object_context_menu.prepend(panel.pt_object_context)
    bpy.types.NODE_MT_context_menu.prepend(panel.change_style_node_menu)

    save_post.append(session._pickle)
    load_post.append(session._load)
    frame_change_pre.append(update_entities)
    render_pre.append(render_pre_handler)

    bpy.types.Scene.MNSession = session.MNSession()  # type: ignore
    bpy.types.Object.uuid = props.uuid_property  # type: ignore
    bpy.types.Object.mn = PointerProperty(type=props.MolecularNodesObjectProperties)  # type: ignore
    bpy.types.Scene.mn = PointerProperty(type=props.MolecularNodesSceneProperties)  # type: ignore
    bpy.types.Object.mn_trajectory_selections = CollectionProperty(  # type: ignore
        type=props.TrajectorySelectionItem  # type: ignore
    )
    # bpy.types.Object.mn_annotations is dynamically created and updated based
    # on different annotation types. It has to be a top level property to avoid
    # AttributeError: '_PropertyDeferred' object has no attribute '...'

    _is_registered = True


def unregister():
    global _is_registered

    for op in all_classes:
        try:
            bpy.utils.unregister_class(op)
        except Exception as e:
            print(e)
            pass

    bpy.types.NODE_MT_add.remove(node_menu.add_node_menu)
    bpy.types.VIEW3D_MT_object_context_menu.remove(panel.pt_object_context)
    bpy.types.NODE_MT_context_menu.remove(panel.change_style_node_menu)

    save_post.remove(session._pickle)
    load_post.remove(session._load)
    frame_change_pre.remove(update_entities)
    render_pre.remove(render_pre_handler)
    del bpy.types.Scene.MNSession  # type: ignore
    del bpy.types.Scene.mn  # type: ignore
    del bpy.types.Object.mn  # type: ignore
    del bpy.types.Object.mn_trajectory_selections  # type: ignore
    del bpy.types.Object.mn_annotations  # type: ignore

    _is_registered = False



import bpy
import MDAnalysis as mda
import numpy as np
import pytest
from databpy.object import LinkedObjectError
import molecularnodes as mn
from molecularnodes.entities.trajectory import oxdna
from .constants import data_dir

pytestmark = pytest.mark.filterwarnings(
    "ignore:.*no reference attributes.*:UserWarning"
)


class TestOXDNAReading:
    @pytest.fixture(scope="module")
    def file_holl_top(self):
        return data_dir / "oxdna/holliday.top"

    @pytest.fixture(scope="module")
    def file_top_new(self):
        return data_dir / "oxdna/top_new.top"

    @pytest.fixture(scope="module")
    def file_top_new_custom(self):
        return data_dir / "oxdna/top_new_custom.top"

    @pytest.fixture(scope="module")
    def file_top_old(self):
        return data_dir / "oxdna/top_old.top"

    @pytest.fixture(scope="module")
    def file_traj_old_new(self):
        return data_dir / "oxdna/traj_old_new.dat"

    @pytest.fixture(scope="module")
    def file_holl_dat(self):
        return data_dir / "oxdna/holliday.dat"

    @pytest.fixture(scope="module")
    def universe(self, file_holl_top, file_holl_dat):
        return mda.Universe(
            file_holl_top,
            file_holl_dat,
            format=oxdna.OXDNAReader,
            topology_format=oxdna.OXDNAParser,
        )

    def test_read_as_universe(self, file_holl_top, file_holl_dat):
        u = mda.Universe(
            file_holl_top,
            file_holl_dat,
            format=oxdna.OXDNAReader,
            topology_format=oxdna.OXDNAParser,
        )
        assert u.atoms.n_atoms == 98

    def test_univ_as_traj(self, universe):
        traj = oxdna.OXDNA(universe, create_object=False)
        assert traj.universe
        with pytest.raises(LinkedObjectError):
            traj.object
        assert all([x in ["A", "C", "T", "G"] for x in traj.atoms.resnames])

    def test_univ_snapshot(self, universe: mda.Universe, snapshot_custom):
        traj = oxdna.OXDNA(universe)
        for name in ["position", "res_name", "res_id", "chain_id"]:
            assert snapshot_custom == str(traj[name])

    def test_detect_new_top(self, file_top_old, file_top_new, file_top_new_custom):
        assert oxdna.OXDNAParser._is_new_topology(file_top_new)
        assert oxdna.OXDNAParser._is_new_topology(file_top_new_custom)
        assert not oxdna.OXDNAParser._is_new_topology(file_top_old)

    def test_topo_reading(
        self,
        file_top_old,
        file_top_new,
        file_top_new_custom,
    ):
        top_new = oxdna.OXDNAParser._read_topo_new(file_top_new)
        top_new_custom = oxdna.OXDNAParser._read_topo_new(file_top_new_custom)
        top_old = oxdna.OXDNAParser._read_topo_old(file_top_old)

        for top in [top_new, top_old, top_new_custom]:
            assert top.n_atoms == 12
            assert top.n_residues == 12

    @pytest.mark.parametrize("topfile", ["top_new", "top_new_custom", "top_old"])
    def test_comparing_topologies(self, snapshot, topfile, file_traj_old_new):
        u = mda.Universe(
            data_dir / f"oxdna/{topfile}.top",
            file_traj_old_new,
            topology_format=oxdna.OXDNAParser,
            format=oxdna.OXDNAReader,
        )
        traj = oxdna.OXDNA(u)
        assert len(traj) == 12
        assert snapshot == traj.atoms.bonds.indices.tolist()
        for att in ["res_id", "chain_id", "res_name"]:
            assert snapshot == str(traj[att])

    def test_reading_example(self):
        traj = oxdna.OXDNA(
            mda.Universe(
                data_dir / "CanDo2oxDNA/top.top",
                data_dir / "CanDo2oxDNA/traj.oxdna",
                topology_format=oxdna.OXDNAParser,
                format=oxdna.OXDNAReader,
            )
        )
        assert len(np.unique(traj.named_attribute("res_id"))) == 15166
        assert len(np.unique(traj.named_attribute("chain_id"))) == 178

    def test_session_register(self, file_holl_top, file_holl_dat):
        session = mn.session.get_session()
        u = mda.Universe(
            file_holl_top,
            file_holl_dat,
            topology_format=oxdna.OXDNAParser,
            format=oxdna.OXDNAReader,
        )
        traj = oxdna.OXDNA(u)

        assert isinstance(session.get(traj.uuid), oxdna.OXDNA)
        assert traj._mn_entity_type == mn.entities.base.EntityType.MD_OXDNA.value

    def test_reload_lost_connection(self, snapshot, file_holl_top, file_holl_dat):
        session = mn.session.get_session()
        u = mda.Universe(
            file_holl_top,
            file_holl_dat,
            topology_format=oxdna.OXDNAParser,
            format=oxdna.OXDNAReader,
        )
        traj = oxdna.OXDNA(u)
        obj_name = traj.name
        bpy.context.scene.frame_set(1)
        pos1 = traj.position
        bpy.context.scene.frame_set(2)
        pos2 = traj.position
        assert not np.allclose(pos1, pos2)

        traj_old = session.entities.pop(traj.uuid)

        # the position shouldn't change as we have removed the traj from the session
        bpy.context.scene.frame_set(3)
        pos3 = traj.position
        assert np.allclose(pos2, pos3)
        del traj

        bpy.data.objects[obj_name].select_set(True)
        bpy.ops.mn.reload_trajectory()

        # when reloading the object, a brand new traj had to be created, which updates
        # the uuid on the object, so the old traj will not longer be able to find any
        # matching object and instead we'll have to look back up a new traj based on the
        # the object's uuid
        with pytest.raises(LinkedObjectError):
            traj_old.object.name

        traj = bpy.context.scene.MNSession.get(bpy.context.active_object.uuid)
        assert traj is not None

        pos3 = traj.position

        assert not np.allclose(pos2, pos3)



import bpy
import pytest
from numpy.testing import assert_allclose
import molecularnodes as mn


def test_setting_material():
    mol = mn.Molecule.fetch("4ozs").add_style("cartoon")
    s = mol.styles[0]
    s.material = "MN Default"
    assert isinstance(s.material, mn.material.TreeInterface)
    assert isinstance(s.material.material, bpy.types.Material)
    s.material = mn.material.AmbientOcclusion()
    assert isinstance(s.material, mn.material.TreeInterface)
    assert isinstance(s.material.material, bpy.types.Material)
    with pytest.raises(AttributeError):
        s.material.non_existent_property
    with pytest.raises(AttributeError):
        s.material.non_existent_property = 1.0
    with pytest.raises(mn.nodes.interface.SocketLinkedError):
        s.material.mix_a


def test_setting_material_in_apply_builtins():
    materials = [
        mn.material.AmbientOcclusion(),
        mn.material.Default(),
        mn.material.FlatOutline(),
        mn.material.Squishy(),
        mn.material.TransparentOutline(),
    ]
    for material in materials:
        mol = mn.Molecule.fetch("4ozs").add_style("cartoon", material=material)
        assert mol is not None, f"Failed to apply {material.__class__.__name__}"


def test_setting_material_in_apply_custom():
    def create_new_material(name="New_Material", color=(1, 0, 0, 1)):
        if name in bpy.data.materials:
            material = bpy.data.materials[name]
        else:
            material = bpy.data.materials.new(name=name)
        material.use_nodes = True
        material.node_tree.nodes.clear()
        principled_node = material.node_tree.nodes.new("ShaderNodeBsdfPrincipled")
        principled_node.location = (0, 0)
        principled_node.inputs["Base Color"].default_value = color
        principled_node.inputs["Metallic"].default_value = 0.5
        principled_node.inputs["Roughness"].default_value = 0.2
        output_node = material.node_tree.nodes.new("ShaderNodeOutputMaterial")
        output_node.location = (300, 0)
        material.node_tree.links.new(
            principled_node.outputs["BSDF"], output_node.inputs["Surface"]
        )
        return material

    new_mat = create_new_material(name="My_Red_Material", color=(1, 0, 0, 1))
    mol = mn.Molecule.fetch("4ozs").add_style("cartoon", material=new_mat)
    assert mol is not None


def test_material_creator():
    pearl_mat = mn.material.Material.pearl()
    assert isinstance(pearl_mat, bpy.types.Material)


def test_generic_material():
    mol = mn.Molecule.fetch("4ozs").add_style("cartoon")
    s = mol.styles[0]
    s.material = bpy.data.materials["Material"]
    assert s.material.material.name == "Material"  # type: ignore
    with pytest.raises(KeyError):
        s.material = "Non-existent material"
    with pytest.raises(TypeError):
        s.material = 1
    assert isinstance(s.material, mn.material.TreeInterface)
    assert isinstance(s.material.material, bpy.types.Material)

    assert_allclose(s.material.principled_bsdf_base_color, (0.8, 0.8, 0.8, 1.0))
    s.material = "MN Ambient Occlusion"
    assert s.material.material.name == "MN Ambient Occlusion"  # type: ignore


def test_ambient_occlusion():
    mol = mn.Molecule.fetch("4ozs").add_style("cartoon")
    s = mol.styles[0]
    assert isinstance(s.material, mn.material.TreeInterface)
    assert isinstance(s.material.material, bpy.types.Material)
    s.material = "MN Ambient Occlusion"
    assert s.material.material.name == "MN Ambient Occlusion"

    assert_allclose(s.material.ambient_occlusion_distance, 1.0)
    s.material.ambient_occlusion_distance = 0.1
    assert_allclose(s.material.ambient_occlusion_distance, 0.1)


def test_material_attribute_access():
    for material in [
        mn.material.AmbientOcclusion,
        mn.material.Default,
        mn.material.FlatOutline,
        mn.material.Squishy,
        mn.nodes.material.dynamic_material_interface(bpy.data.materials["Material"]),
    ]:
        for attr in material.__dict__:
            if not attr.startswith("_"):
                assert hasattr(material, attr)
                assert getattr(material, attr) is not None



import numpy as np
import pytest
import molecularnodes as mn
from .emdb_pooch import fetch_emdb_map


@pytest.fixture
def emdb_density_map():
    file = fetch_emdb_map("EMD-48397")
    # we don't just use Path().stem because usually is .map.gz
    base = file.name.split(".")[0]
    vdb_path = file.parent / f"{base}.vdb"
    vdb_path.unlink(missing_ok=True)
    return file


def test_emdb_api_density_load(emdb_density_map):
    density = mn.entities.density.load(emdb_density_map)
    pos = density.named_attribute("position")

    # Basic sanity checks only; remote data can change slightly over time
    assert len(pos) > 1000
    avg = np.mean(pos, axis=0)
    assert np.linalg.norm(avg) > 0.1

    assert density.object.mn.entity_type == "density"
    assert density.object.users_collection[0] == mn.blender.coll.mn()



from ..annotations.base import BaseAnnotation


class Label2D(BaseAnnotation):
    """
    Show a label in 2D space (viewport / render)

    Attributes
    ----------
    text: str
        Text to be displayed

    location: tuple[float, float]
        Normalized coordinates (0.0 - 1.0) of the postion in viewport / render

    """

    annotation_type = "label_2d"

    text: str
    location: tuple[float, float] = (0.5, 0.5)

    def validate(self) -> bool:
        params = self.interface
        x, y = params.location
        if (not 0 <= x <= 1) or (not 0 <= y <= 1):
            raise ValueError("Normalized coordinates should lie between 0 and 1")
        return True

    def draw(self) -> None:
        params = self.interface
        self.draw_text_2d_norm(params.location, params.text)


class Label3D(BaseAnnotation):
    """
    Show a label in 3D space (world space)

    Attributes
    ----------
    text: str
        Text to be displayed

    location: tuple[float, float, float]
        3D Coordinates of the location to display text

    """

    annotation_type = "label_3d"

    text: str
    location: tuple[float, float, float] = (0.0, 0.0, 0.0)

    def draw(self) -> None:
        params = self.interface
        self.draw_text_3d(params.location, params.text)



from pathlib import Path
from .grids import Grids


def load(
    file_path: str | Path,
    name: str = "NewDensity",
    invert: bool = False,
    setup_nodes: bool = True,
    style: str = "density_surface",
    center: bool = False,
    overwrite: bool = False,
):
    density = Grids(
        file_path=file_path, center=center, invert=invert, overwrite=overwrite
    )
    density.create_object(
        name=Path(file_path).name, setup_nodes=setup_nodes, style=style
    )
    return density



import bpy


def annotations_node_tree():
    node_group = bpy.data.node_groups.new(type="GeometryNodeTree", name="Annotations")
    node_group.is_modifier = True

    # Output Socket Geometry
    node_group.interface.new_socket(
        name="Geometry", in_out="OUTPUT", socket_type="NodeSocketGeometry"
    )

    # Input Socket Geometry
    node_group.interface.new_socket(
        name="Geometry", in_out="INPUT", socket_type="NodeSocketGeometry"
    )

    # initialize nodes
    # node Group Input
    group_input = node_group.nodes.new("NodeGroupInput")
    group_input.name = "Group Input"

    # node Group Output
    group_output = node_group.nodes.new("NodeGroupOutput")
    group_output.name = "Group Output"

    # node Mesh to Curve
    mesh_to_curve = node_group.nodes.new("GeometryNodeMeshToCurve")
    mesh_to_curve.name = "Mesh to Curve"

    # node Set Curve Radius
    set_curve_radius = node_group.nodes.new("GeometryNodeSetCurveRadius")
    set_curve_radius.name = "Set Curve Radius"

    # node Is Line Attribute
    is_line_attribute = node_group.nodes.new("GeometryNodeInputNamedAttribute")
    is_line_attribute.name = "Is Line Attribute"
    is_line_attribute.data_type = "BOOLEAN"
    is_line_attribute.inputs["Name"].default_value = "is_line"

    # node Thickness Attribute
    thickness_attribute = node_group.nodes.new("GeometryNodeInputNamedAttribute")
    thickness_attribute.name = "Thickness Attribute"
    thickness_attribute.inputs["Name"].default_value = "thickness"

    # node Curve Circle
    curve_circle = node_group.nodes.new("GeometryNodeCurvePrimitiveCircle")
    curve_circle.name = "Curve Circle"
    curve_circle.inputs["Resolution"].default_value = 32
    curve_circle.inputs["Radius"].default_value = 0.001

    # node Curve to Mesh
    curve_to_mesh = node_group.nodes.new("GeometryNodeCurveToMesh")
    curve_to_mesh.name = "Curve to Mesh"
    curve_to_mesh.inputs["Fill Caps"].default_value = True

    # node Join Geometry
    join_geometry = node_group.nodes.new("GeometryNodeJoinGeometry")
    join_geometry.name = "Join Geometry"

    # node Set Material Index
    set_material_index = node_group.nodes.new("GeometryNodeSetMaterialIndex")
    set_material_index.name = "Set Material Index"

    # node Material Slot Attribute
    material_slot_attribute = node_group.nodes.new("GeometryNodeInputNamedAttribute")
    material_slot_attribute.name = "Material Slot Attribute"
    material_slot_attribute.data_type = "INT"
    # Name
    material_slot_attribute.inputs["Name"].default_value = "material_slot_index"

    # node Shade Smooth Attribute
    shade_smooth_attribute = node_group.nodes.new("GeometryNodeInputNamedAttribute")
    shade_smooth_attribute.name = "Shade Smooth Attribute"
    shade_smooth_attribute.data_type = "BOOLEAN"
    # Name
    shade_smooth_attribute.inputs["Name"].default_value = "shade_smooth"

    # node Set Shade Smooth
    set_shade_smooth = node_group.nodes.new("GeometryNodeSetShadeSmooth")
    set_shade_smooth.name = "Set Shade Smooth"

    # Set locations
    group_input.location = (-520.0, 0.0)
    group_output.location = (920.0, 0.0)
    mesh_to_curve.location = (-160.0, 0.0)
    set_curve_radius.location = (20.0, 0.0)
    is_line_attribute.location = (-340.0, -120.0)
    thickness_attribute.location = (-160.0, -120.0)
    curve_circle.location = (20.0, -140.0)
    curve_to_mesh.location = (200.0, 0.0)
    join_geometry.location = (380.0, 0.0)
    set_material_index.location = (560.0, 0.0)
    material_slot_attribute.location = (380.0, -120.0)
    shade_smooth_attribute.location = (560.0, -140.0)
    set_shade_smooth.location = (740.0, 0.0)

    # initialize links
    # is_line_attribute.Attribute -> set_shade_smooth.Selection
    node_group.links.new(
        shade_smooth_attribute.outputs["Attribute"],
        set_shade_smooth.inputs["Shade Smooth"],
    )
    # set_material_index.Geometry -> set_shade_smooth.Geometry
    node_group.links.new(
        set_material_index.outputs["Geometry"], set_shade_smooth.inputs["Geometry"]
    )
    # set_shade_smooth.Geometry -> group_output.Geometry
    node_group.links.new(
        set_shade_smooth.outputs["Geometry"], group_output.inputs["Geometry"]
    )
    # join_geometry.Geometry -> set_material_index.Geometry
    node_group.links.new(
        join_geometry.outputs["Geometry"], set_material_index.inputs["Geometry"]
    )
    # material_slot_attribute.Attribute -> set_material_index.Material Index
    node_group.links.new(
        material_slot_attribute.outputs["Attribute"],
        set_material_index.inputs["Material Index"],
    )
    # group_input.Geometry -> mesh_to_curve.Mesh
    node_group.links.new(group_input.outputs["Geometry"], mesh_to_curve.inputs["Mesh"])
    if bpy.app.version >= (4, 5, 0):
        # From Blender 4.5.x onwards, need to use Scale input of Curve to Mesh
        # remove set_curve_radius node
        node_group.nodes.remove(set_curve_radius)
        # mesh_to_curve.Curve -> curve_to_mesh.Curve
        node_group.links.new(
            mesh_to_curve.outputs["Curve"], curve_to_mesh.inputs["Curve"]
        )
        # thickness_attribute.Attribute -> curve_to_mesh.Scale
        node_group.links.new(
            thickness_attribute.outputs["Attribute"], curve_to_mesh.inputs["Scale"]
        )
    else:
        # mesh_to_curve.Curve -> set_curve_radius.Curve
        node_group.links.new(
            mesh_to_curve.outputs["Curve"], set_curve_radius.inputs["Curve"]
        )
        # thickness_attribute.Attribute -> set_curve_radius.Radius
        node_group.links.new(
            thickness_attribute.outputs["Attribute"], set_curve_radius.inputs["Radius"]
        )
        # set_curve_radius.Curve -> curve_to_mesh.Curve
        node_group.links.new(
            set_curve_radius.outputs["Curve"], curve_to_mesh.inputs["Curve"]
        )
    # is_line_attribute.Attribute -> mesh_to_curve.Selection
    node_group.links.new(
        is_line_attribute.outputs["Attribute"], mesh_to_curve.inputs["Selection"]
    )
    # curve_circle.Curve -> curve_to_mesh.Profile Curve
    node_group.links.new(
        curve_circle.outputs["Curve"], curve_to_mesh.inputs["Profile Curve"]
    )
    # curve_to_mesh.Mesh -> join_geometry.Geometry
    node_group.links.new(
        curve_to_mesh.outputs["Mesh"], join_geometry.inputs["Geometry"]
    )
    # group_input.Geometry -> join_geometry.Geometry
    node_group.links.new(
        group_input.outputs["Geometry"], join_geometry.inputs["Geometry"]
    )
    return node_group



from biotite.structure.io.mol import MOLFile
from .reader import ReaderBase


class SDFReader(ReaderBase):
    _extra_annotations = {}

    def read(self, file_path):
        return MOLFile.read(file_path)

    def get_structure(self):
        return self.file.get_structure()

    def _assemblies(self):
        # TODO maybe look into symmetry operations for small mols
        return None



import itertools
import MDAnalysis as mda
import numpy as np
import pytest
import molecularnodes as mn
from .constants import codes, data_dir


@pytest.mark.parametrize(
    "code, format", itertools.product(codes, ["bcif", "cif", "pdb"])
)
def test_biotite_converter(code, format):
    # fetch and load molecule through biotite
    mol = mn.Molecule.fetch(code, format=format, cache=data_dir)
    # create MDAnalysis universe from biotite structure
    u = mda.Universe(mn.converters.BiotiteWrapper(mol.array))
    # load trajectory into Blender
    traj = mn.Trajectory(u)
    # check coords
    assert np.allclose(mol.position, traj.position)
    # check named attributes
    attrs = [
        "chain_id",
        "res_id",
        "res_name",
        "atom_name",
        "atom_id",
        "b_factor",
        "occupancy",
        "charge",
    ]
    for attr in attrs:
        assert np.allclose(mol.named_attribute(attr), traj.named_attribute(attr))
    # check string attributes
    assert np.array_equal(mol.array.element, traj.atoms.elements)
    assert np.array_equal(mol.array.ins_code, traj.atoms.icodes)
    # check computed attributes
    # the ones that currently differ are commented out with reasons
    computed_attrs = [
        ("mass", 1e-3),
        # ("atomic_number", 0),  # unsure which is correct, computation different
        # ("vdw_radii", 0), # np.char.title in Molecule causes difference
        ("charge", 0),
        ("is_alpha_carbon", 0),
        ("is_solvent", 0),
        # ("is_backbone", 0),  # computation different
        ("is_nucleic", 0),
        ("is_peptide", 0),
        # ("ures_id", 0),  # not currently in Trajectory
        # ("lipophobicity", 0),  # not currently in Trajectory
        # ("Color", 0),  # not currently in Trajectory
        # ("is_hetero", 0),  # not currently in Trajectory
        # ("is_side_chain", 0),  # not currently in Trajectory
        # ("is_carb", 0),  # not currently in Trajectory
        # ("sec_struct", 0),  # not currently in Trajectory
        # ("entity_id", 0),  # not currently in Trajectory
        # ("asym_id", 0),  # not currently in Trajectory
        # ("pdb_model_num", 0),  # not currently in Trajectory
    ]
    for attr, rtol in computed_attrs:
        assert np.allclose(
            mol.named_attribute(attr), traj.named_attribute(attr), rtol=rtol
        )



import pathlib
import bpy
import nodepad
import molecularnodes as mn

DOCS_FOLDER = pathlib.Path(__file__).resolve().parent

# load the data file which contains all of the nodes to build docs for
bpy.ops.wm.open_mainfile(filepath=str(mn.assets.MN_DATA_FILE))


header = """---
toc: true
toc-depth: 2
fig-align: center
---
"""

# generate all of the node related docs
for submenu in mn.ui.node_menu.menu_items.submenus:
    with open(DOCS_FOLDER / f"nodes/{submenu.name}.qmd", "w") as file:
        file.write(header)
        file.write(f"# {submenu.title}\n\n")
        if submenu.description:
            file.write(submenu.description)
            file.write("\n\n")
        for menu_item in submenu.items:
            if menu_item.is_break:
                continue
            if menu_item.backup is not None:
                name = menu_item.backup
            else:
                name = menu_item.name
            doc = nodepad.Documenter(menu_item.tree)
            try:
                doc.lookup_info(menu_item.to_dict())
            except AttributeError as e:
                print(e)

            if menu_item.description != "":
                doc.description += "\n\n" + menu_item.description

            file.write(doc.as_markdown())
            file.write("\n\n")


# write the data table page
with open(DOCS_FOLDER / "data_table.qmd", "w") as file:
    file.write(header)
    file.write("# Data Tables\n\n")
    file.write(
        "The different lookup tables that are used to conver strings to integers in Molecular Nodes.\n\n"
        "Code for this can be found on the [GitHub Page](https://github.com/BradyAJohnston/MolecularNodes/blob/main/molecularnodes/assets/data.py)\n\n"
    )
    file.write("### Residue Names\n\n| Name | Integer |\n|----------:|:------------|\n")
    for name, res in mn.assets.data.residues.items():
        file.write(f"| {name} | `{res['res_name_num']}::Int` |\n")
    file.write("\n")
    file.write("\n")

    file.write("### Atom Names\n\n| Name | Integer |\n|----------:|:------------|\n")
    for name, value in mn.assets.data.atom_names.items():
        file.write(f"| {name} | `{value}::Int` |\n")
    file.write("\n")



import sys
import pytest

argv = sys.argv
argv = argv[argv.index("--") + 1 :]


# run this script like this:
# /Applications/Blender.app/Contents/MacOS/Blender -b -P tests/run.py -- . -v
# /Applications/Blender.app/Contents/MacOS/Blender -b -P tests/run.py -- . -k test_color_lookup_supplied


def main():
    # bpy.ops.preferences.addon_enable(module="bl_ext.user_default.molecularnodes")
    # bpy.ops.wm.save_userpref()
    # run the test suite, and we have to manually return the result value if non-zero
    # value is returned for a failing test
    if len(argv) == 0:
        result = pytest.main()
    else:
        result = pytest.main(argv)
    if result.value != 0:
        sys.exit(result.value)


if __name__ == "__main__":
    main()



from typing import Type, TypeVar
import bpy
import numpy as np
from mathutils import Vector
from .arrange import arrange_tree

NODE_SPACING = 250


class TreeInterface:
    def __init__(self):
        self._dynamic_properties = set()
        self.tree: bpy.types.NodeTree
        self._allowable_properties = {"tree", "remove"}

    def __getattr__(self, name):
        # This method is only called when the attribute doesn't exist
        # No need to check if it exists in _dynamic_properties
        raise AttributeError(
            f"'{self.__class__.__name__}' has no attribute '{name}'.\n"
            f"Potentially a dynamic property in {self._dynamic_properties}"
        )

    def __setattr__(self, name, value):
        # For setting attributes, we need to check if it's allowed
        if (
            name != "_dynamic_properties"
            and not name.startswith("_")
            and hasattr(self, "_dynamic_properties")
            and name not in self._dynamic_properties
            and name not in self._allowable_properties
        ):
            raise AttributeError(
                f"Cannot set non-existent property '{name}' on '{self.__class__.__name__}' "
                f"\nAvailable dynamic properties: {self._dynamic_properties}"
            )
        # Use the normal attribute setting mechanism
        super().__setattr__(name, value)
        # update view layer to reflect changes
        bpy.context.view_layer.update()

    def _register_property(self, name):
        """Register a dynamic property that can be set on this interface"""
        self._dynamic_properties.add(name)


T = TypeVar(
    "T",
    float,
    int,
    bool,
    str,
    np.ndarray,
    tuple[float, float, float],
    list[float, float, float],
)

# Named Attribute node only allows the following data types
named_attribute_data_types = [
    "FLOAT",
    "INT",
    "FLOAT_VECTOR",
    "FLOAT_COLOR",
    "BOOLEAN",
    "QUATERNION",
    "FLOAT4X4",
]


class SocketLinkedError(Exception):
    def __init__(
        self, message="Socket is linked, default value has no bearing on the result"
    ):
        self.message = message
        super().__init__(self.message)


def input_named_attribute(
    socket: bpy.types.NodeSocket, name: str, data_type: str | None
) -> bpy.types.GeometryNode:
    """
    Add a named attribute node to the tree and connect it to the given socket
    """
    tree = socket.node.id_data
    node_na = tree.nodes.new("GeometryNodeInputNamedAttribute")

    if data_type is not None:
        node_na.data_type = data_type
    node_na.inputs["Name"].default_value = name
    node_na.location = socket.node.location - Vector([NODE_SPACING, 0])

    tree.links.new(node_na.outputs["Attribute"], socket)

    return node_na


def _socket_bool(socket: bpy.types.NodeSocket):
    def getter(self) -> bool:
        if socket.is_linked:
            raise SocketLinkedError()
        else:
            return getattr(socket, "default_value")

    def setter(self, value: bool | str) -> None:
        if socket.is_linked:
            socket.node.id_data.links.remove(socket.links[0])
            arrange_tree(socket.node.id_data)
        if isinstance(value, bool):
            setattr(socket, "default_value", value)

        if isinstance(value, str):
            interface_item = socket.node.node_tree.interface.items_tree[socket.name]
            if interface_item.force_non_field:
                raise ValueError(
                    f"Cannot use a named attribute for this input: '{socket}' as it accepts only single values"
                )
            input_named_attribute(socket, value, "BOOLEAN")
            arrange_tree(socket.node.id_data)

    return property(getter, setter)


def socket_lookup(node_name: str, socket_name: str, type_: type[T]):
    def getter(self) -> type[T]:
        return self.tree.nodes[node_name].inputs[socket_name].default_value

    def setter(self, value: T) -> None:
        self.tree.nodes[node_name].inputs[socket_name].default_value = value

    return property(getter, setter)


def check_linked(socket: bpy.types.NodeSocket):
    if socket.is_linked:
        raise SocketLinkedError()


def remove_linked(socket: bpy.types.NodeSocket):
    if socket.is_linked:
        socket.node.id_data.links.remove(socket.links[0])
        arrange_tree(socket.node.id_data)


def create_socket_property(
    socket: bpy.types.NodeSocket,
    value_type: Type,
    data_type: str | None = None,
    length_check: int | None = None,
):
    """
    Generic function to create socket properties with consistent behavior.

    Parameters:
    - socket: The socket to create a property for
    - value_type: The expected Python type(s) for the value
    - data_type: The Blender data type for named attributes
    - length_check: If not None, validates that array-like values have this length
    """

    def getter(self):
        check_linked(socket)
        return getattr(socket, "default_value")

    def setter(self, value):
        remove_linked(socket)

        if isinstance(value, str) and data_type in named_attribute_data_types:
            input_named_attribute(socket, value, data_type)
        elif isinstance(value, value_type):
            # For array-like types, check length if specified
            if length_check is not None and hasattr(value, "__len__"):
                if len(value) != length_check:
                    raise ValueError(
                        f"Value must have length {length_check}, not {len(value)}"
                    )
            setattr(socket, "default_value", value)
        else:
            raise ValueError(
                f"Value must be a {value_type.__name__}, not a {type(value)}"
            )

    return property(getter, setter, doc=socket.description)


def getset_float(socket: bpy.types.NodeSocketFloat):
    return create_socket_property(socket, (float, int), "FLOAT")


def getset_int(socket: bpy.types.NodeSocketInt):
    return create_socket_property(socket, int, "INT")


def getset_bool(socket: bpy.types.NodeSocketBool):
    return create_socket_property(socket, bool, "BOOLEAN")


def getset_vector(socket: bpy.types.NodeSocket):
    return create_socket_property(socket, (np.ndarray, list, tuple), "FLOAT_VECTOR", 3)


def getset_menu(socket: bpy.types.NodeSocketMenu):
    def getter(self) -> str:
        check_linked(socket)
        return getattr(socket, "default_value")

    def setter(self, value: str):
        if isinstance(value, str):
            remove_linked(socket)
            # try first to use title case so we can supply lowercase values
            # but if that fails then try and fail with the orignal value
            try:
                setattr(socket, "default_value", value.title())
            except TypeError:
                setattr(socket, "default_value", value)
        else:
            raise ValueError(f"Value must be a string, not a {type(value)}")

    return property(getter, setter, doc=socket.description)


def getset_string(socket: bpy.types.NodeSocket):
    return create_socket_property(socket, str, "STRING")


def getset_color(socket: bpy.types.NodeSocketColor):
    return create_socket_property(socket, (np.ndarray, list, tuple), "FLOAT_VECTOR", 4)


def getset_rotation(socket: bpy.types.NodeSocketRotation):
    return create_socket_property(socket, (np.ndarray, list, tuple), "QUATERNION", 4)


def socket(socket: bpy.types.NodeSocket):
    match socket.type:
        case "FLOAT" | "VALUE":
            return getset_float(socket)
        case "INT":
            return getset_int(socket)
        case "BOOLEAN":
            return getset_bool(socket)
        case "ROTATION":
            return getset_rotation(socket)
        case "VECTOR":
            return getset_vector(socket)
        case "STRING":
            return getset_string(socket)
        case "RGBA":
            return getset_color(socket)
        case "MENU":
            return getset_menu(socket)
        case _:
            raise ValueError(f"Unknown socket type: {socket.type}")


def option(node_name: str, input: str, type_: type[T]):
    def getter(self) -> T:
        return self.tree.nodes[node_name][input]

    def setter(self, value: T) -> None:
        setattr(self.tree.nodes[node_name], input, value)

    return property(getter, setter)



import bpy
from bpy.props import BoolProperty, StringProperty
from .. import __package__, assets, download


def addon_preferences(
    context: bpy.types.Context | None = None,
) -> bpy.types.AddonPreferences:
    if context is None:
        context = bpy.context
    try:
        return context.preferences.addons[__package__].preferences
    except KeyError:
        key = "bl_ext.vscode_development.molecularnodes"
        if key in context.preferences.addons:
            return context.preferences.addons[key].preferences
        return None


class MN_OT_Template_Install(bpy.types.Operator):
    bl_idname = "mn.template_install"
    bl_label = "Install Template"
    bl_description = "Install the Molecular Nodes startup template file."

    def execute(self, context):
        assets.template.install()
        self.report({"INFO"}, "Installed Molecular Nodes template.")
        return {"FINISHED"}


class MN_OT_Template_Uninstall(bpy.types.Operator):
    bl_idname = "mn.template_uninstall"
    bl_label = "Uninstall Template"
    bl_description = "Uninstall the Molecular Nodes startup template file."

    @classmethod
    def poll(cls, context):
        return assets.template.is_installed()

    def execute(self, context):
        try:
            assets.template.uninstall()
            self.report({"INFO"}, "Uninstalled Molecular Nodes template.")
        except FileNotFoundError:
            self.report({"WARNING"}, "Template not installed.")

        return {"FINISHED"}


class MolecularNodesPreferences(bpy.types.AddonPreferences):
    bl_idname = __package__

    cache_dir: StringProperty(  # type: ignore
        name="Cache Directory",
        description="Where to store the structures downloaded from the Protein Data Bank",
        default=str(download.CACHE_DIR),
        subtype="DIR_PATH",
    )

    cache_download: BoolProperty(  # type: ignore
        name="Cache Downloads",
        description="Cache downloaded files from the Protein Data Bank",
        default=True,
    )

    def draw(self, context):
        layout = self.layout
        layout.label(
            text="Where and if to store downloaded files for faster subsequent loading:"
        )
        row = layout.row()
        row.prop(self, "cache_download", text="")
        col = row.column()
        col.prop(self, "cache_dir")
        col.enabled = self.cache_download
        layout.label(
            text="Install the Molecular Nodes template file, to start Blender with useful default settings"
        )
        row = layout.row()
        if not assets.template.is_installed():
            text = "Install Template"
        else:
            text = "Reinstall Template"

        row.operator("mn.template_install", text=text)
        row.operator("mn.template_uninstall")


CLASSES = [MN_OT_Template_Install, MN_OT_Template_Uninstall, MolecularNodesPreferences]



import itertools
import numpy as np
import pytest
import molecularnodes as mn
from .constants import attributes, codes, data_dir

formats = ["pdb", "cif", "bcif"]


@pytest.mark.parametrize("code, format", itertools.product(codes, formats))
def test_attribute(snapshot_custom, code, format):
    mol = mn.Molecule.fetch(code, cache=data_dir, format=format)
    for attribute in attributes:
        try:
            print(f"{attribute=}")
            assert snapshot_custom == mol.named_attribute(attribute)
        except AttributeError as e:
            assert snapshot_custom == e


def test_store_named_attribute(snapshot_custom):
    mol = mn.Molecule.fetch("8H1B", cache=data_dir, format="bcif")
    before = mol.named_attribute("position")
    mol.store_named_attribute(mol.named_attribute("position") + 10, "position")
    after = mol.named_attribute("position")

    assert not np.allclose(before, after)


def test_uv_map(snapshot_custom):
    mol = mn.Molecule.fetch("1cd3", cache=data_dir, format="bcif").add_style("ribbon")
    mol.styles[0].uv_map = True
    mol.styles[0].quality = 1
    assert snapshot_custom == mol.named_attribute("uv_map", evaluate=True)[:1000]
    assert snapshot_custom == mol.named_attribute("uv_map", evaluate=True)[-1000:]


def test_bond_attributes(snapshot_custom):
    mol = mn.Molecule.fetch("1bna", cache=data_dir, format="bcif").add_style(
        mn.StyleBallAndStick(sphere_geometry="Mesh"),
    )

    for attr in mol.list_attributes(evaluate=True, drop_hidden=True):
        assert snapshot_custom == mol.named_attribute(attr, evaluate=True)



import gzip
import io
import os
import tempfile
import time
from pathlib import Path
from unittest.mock import Mock, patch
import biotite.database.rcsb as rcsb
import pytest
import requests
from biotite.structure.io import load_structure
from molecularnodes.download import (
    CACHE_DIR,
    FileDownloadPDBError,
    StructureDownloader,
)
from .constants import codes

# currently can't figure out downloading from other services
DATABASES = ["rcsb"]
SLEEP_TIME = 2.0


def _filestart(format):
    if format == "cif":
        return "data_"
    else:
        return "HEADER"


def test_download_raises_error_on_invalid_format():
    downloader = StructureDownloader()
    with pytest.raises(ValueError) as excinfo:
        downloader.download("1abc", "invalid_format")
    assert (
        "File format 'invalid_format' not in: supported_formats=['cif', 'pdb', 'bcif']"
        in str(excinfo.value)
    )
    time.sleep(SLEEP_TIME)


def test_fail_download_pdb_large_structure_raises():
    downloader = StructureDownloader()
    with pytest.raises(FileDownloadPDBError):
        downloader.download("7D6Z", format="pdb")
    time.sleep(SLEEP_TIME)


@pytest.mark.parametrize("format", ["cif", "bcif", "pdb"])
def test_compare_biotite(format):
    with tempfile.TemporaryDirectory() as temp_dir:
        downloader = StructureDownloader(cache=temp_dir)
        struc_download = load_structure(downloader.download("4ozs", format=format))

        with tempfile.TemporaryDirectory() as biotite_temp_dir:
            struc_biotite = load_structure(
                rcsb.fetch("4ozs", format=format, target_path=biotite_temp_dir)
            )
            assert struc_download == struc_biotite
            time.sleep(SLEEP_TIME)


@pytest.mark.parametrize("code", codes)
@pytest.mark.parametrize("database", DATABASES)
@pytest.mark.parametrize("format", ["pdb", "cif", "bcif"])
def test_fetch_with_cache(tmpdir, code, format, database):
    cache_dir = tmpdir.mkdir("cache")
    downloader = StructureDownloader(cache=str(cache_dir))
    file = downloader.download(code, format, database=database)

    assert isinstance(file, Path)
    assert os.path.isfile(file)
    assert file.name == f"{code}.{format}"

    if format == "bcif":
        # Binary format needs to be opened in binary mode
        with open(file, "rb") as f:
            content = f.read()
        # Use the binary file starts defined in test_fetch_with_binary_format
        assert content.startswith(b"\x83\xa7")
    else:
        # Text formats can be opened normally
        with open(file, "r") as f:
            content = f.read()
        assert content.startswith(_filestart(format))
    time.sleep(SLEEP_TIME)


@pytest.mark.parametrize("code", codes)
@pytest.mark.parametrize("database", DATABASES)
@pytest.mark.parametrize("format", ["pdb", "cif", "bcif"])
def test_fetch_new_code(tmpdir, code, format, database):
    cache_dir = tmpdir.mkdir("cache")
    downloader = StructureDownloader(cache=str(cache_dir))
    code = f"pdb_{code.upper().rjust(8, '0')}"
    if format == "pdb":
        with pytest.raises(ValueError):
            downloader.download(code, format, database=database)
        return
    if format == "bcif":
        with pytest.raises(FileDownloadPDBError):
            downloader.download(code, format, database=database)
        return
    file = downloader.download(code, format, database=database)

    assert isinstance(file, Path)
    assert os.path.isfile(file)
    assert file.name == f"{code}.{format}"

    with open(file, "r") as f:
        content = f.read()
    assert content.startswith(_filestart(format))
    time.sleep(SLEEP_TIME)


DATABASES = ["rcsb"]  # currently can't figure out downloading from the pdbe


@pytest.mark.parametrize("database", DATABASES)
def test_fetch_with_invalid_format(database):
    code = "4OZS"
    format = "xyz"
    downloader = StructureDownloader(cache=None)

    with pytest.raises(ValueError):
        downloader.download(code, format, database=database)
    time.sleep(SLEEP_TIME)


# tests are failling intermittently - just ignoring for now TODO: stop ignoring
# limit to just 1 download test TODO: use API key to up requests in tests
# @pytest.mark.parametrize("format", ("cif",))
# # @pytest.mark.parametrize("format", ("pdb", "cif", "bcif"))
# @pytest.mark.parametrize(
#     "code",
#     (
#         "K4PA18",
#         # "G1JSI4",
#     ),
# )
# def test_fetch_alphafold(format: str, code: str, tmpdir) -> None:
#     time.sleep(SLEEP_TIME)
#     mol = mn.Molecule.fetch(code, format=format, database="alphafold", cache=tmpdir)
#     assert mol.array


# Test StructureDownloader initialization
class TestStructureDownloaderInit:
    def test_init_with_default_cache(self):
        downloader = StructureDownloader()
        assert downloader.cache == CACHE_DIR
        assert downloader.cache.exists()

    def test_init_with_custom_cache_string(self, tmpdir):
        cache_path = str(tmpdir.mkdir("custom_cache"))
        downloader = StructureDownloader(cache=cache_path)
        assert downloader.cache == Path(cache_path).absolute()
        assert downloader.cache.exists()

    def test_init_with_custom_cache_path(self, tmpdir):
        cache_path = tmpdir.mkdir("custom_cache")
        downloader = StructureDownloader(cache=Path(cache_path))
        assert downloader.cache == Path(cache_path).absolute()

    def test_init_with_no_cache(self):
        downloader = StructureDownloader(cache=None)
        assert downloader.cache is None

    def test_init_creates_cache_directory(self, tmpdir):
        nonexistent_path = tmpdir.join("new_cache")
        assert not nonexistent_path.exists()
        downloader = StructureDownloader(cache=str(nonexistent_path))
        assert downloader.cache.exists()


# Test URL generation
class TestUrlGeneration:
    def test_url_rcsb_cif(self):
        downloader = StructureDownloader(cache=None)
        url = downloader._url("1abc", "cif", "rcsb")
        assert url == "https://files.rcsb.org/download/1abc.cif"

    def test_url_rcsb_pdb(self):
        downloader = StructureDownloader(cache=None)
        url = downloader._url("1abc", "pdb", "rcsb")
        assert url == "https://files.rcsb.org/download/1abc.pdb"

    def test_url_rcsb_bcif(self):
        downloader = StructureDownloader(cache=None)
        url = downloader._url("1abc", "bcif", "rcsb")
        assert url == "https://models.rcsb.org/1abc.bcif"

    def test_url_pdb_database(self):
        downloader = StructureDownloader(cache=None)
        url = downloader._url("1abc", "cif", "pdb")
        assert url == "https://files.rcsb.org/download/1abc.cif"

    def test_url_wwpdb_database(self):
        downloader = StructureDownloader(cache=None)
        url = downloader._url("1abc", "cif", "wwpdb")
        assert url == "https://files.rcsb.org/download/1abc.cif"

    def test_url_unsupported_database(self):
        downloader = StructureDownloader(cache=None)
        with pytest.raises(
            ValueError, match="Database unsupported not currently supported"
        ):
            downloader._url("1abc", "cif", "unsupported")

    def test_url_alphafold_database_not_supported(self):
        # AlphaFold is now handled directly in download() using biotite.database.afdb
        # and should not be supported in _url()
        downloader = StructureDownloader(cache=None)
        with pytest.raises(
            ValueError, match="Database alphafold not currently supported"
        ):
            downloader._url("P12345", "cif", "alphafold")


# Test cache behavior
class TestCacheBehavior:
    def test_cache_hit_returns_existing_file(self, tmpdir):
        cache_dir = tmpdir.mkdir("cache")
        existing_file = cache_dir.join("test.cif")
        existing_file.write("existing content")

        downloader = StructureDownloader(cache=str(cache_dir))

        with patch.object(downloader, "_url"), patch("requests.get") as mock_get:
            result = downloader.download("test", "cif")

            # Should return existing file without making HTTP request
            assert isinstance(result, Path)
            assert result.name == "test.cif"
            assert result.read_text() == "existing content"
            mock_get.assert_not_called()

    @patch("requests.get")
    def test_no_cache_returns_string_io(self, mock_get):
        mock_response = Mock()
        mock_response.text = "test content"
        mock_response.content = b"test content"
        mock_get.return_value = mock_response

        downloader = StructureDownloader(cache=None)
        result = downloader.download("test", "cif")

        assert isinstance(result, io.StringIO)
        assert result.getvalue() == "test content"

    @patch("requests.get")
    def test_no_cache_returns_bytes_io_for_binary(self, mock_get):
        mock_response = Mock()
        mock_response.content = b"binary content"
        mock_get.return_value = mock_response

        downloader = StructureDownloader(cache=None)
        result = downloader.download("test", "bcif")

        assert isinstance(result, io.BytesIO)
        assert result.getvalue() == b"binary content"


# Test error handling
class TestErrorHandling:
    @patch("requests.get")
    def test_http_error_raises_file_download_error(self, mock_get):
        mock_get.side_effect = requests.HTTPError("404 Not Found")

        downloader = StructureDownloader(cache=None)
        with pytest.raises(FileDownloadPDBError, match="404 Not Found"):
            downloader.download("nonexistent", "cif")

    def test_new_format_pdb_code_with_pdb_format_raises_error(self):
        downloader = StructureDownloader(cache=None)
        with pytest.raises(
            ValueError,
            match="New format PDB codes .* are not compatible with .pdb format",
        ):
            downloader.download("pdb_00001abc", "pdb")

    @patch("requests.get")
    def test_invalid_content_type_for_binary(self, mock_get):
        mock_response = Mock()
        mock_response.content = "not bytes"  # Should be bytes for binary format
        mock_get.return_value = mock_response

        downloader = StructureDownloader(cache=None)
        with pytest.raises(ValueError, match="Binary content is not bytes"):
            downloader.download("test", "bcif")

    @patch("requests.get")
    def test_invalid_content_type_for_text(self, mock_get):
        mock_response = Mock()
        mock_response.text = b"not string"  # Should be string for text format
        mock_get.return_value = mock_response

        downloader = StructureDownloader(cache=None)
        with pytest.raises(ValueError, match="Text content is not str"):
            downloader.download("test", "cif")


# Test gzip decompression
class TestGzipDecompression:
    @patch("requests.get")
    def test_gzipped_content_is_decompressed(self, mock_get):
        original_content = b"test binary content"
        gzipped_content = gzip.compress(original_content)

        mock_response = Mock()
        mock_response.content = gzipped_content
        mock_get.return_value = mock_response

        downloader = StructureDownloader(cache=None)
        result = downloader.download("test", "bcif")

        assert isinstance(result, io.BytesIO)
        assert result.getvalue() == original_content

    @patch("requests.get")
    def test_non_gzipped_binary_content_unchanged(self, mock_get):
        content = b"test binary content"

        mock_response = Mock()
        mock_response.content = content
        mock_get.return_value = mock_response

        downloader = StructureDownloader(cache=None)
        result = downloader.download("test", "bcif")

        assert isinstance(result, io.BytesIO)
        assert result.getvalue() == content


# Test input sanitization
class TestInputSanitization:
    @patch("requests.get")
    def test_code_with_whitespace_is_stripped(self, mock_get):
        mock_response = Mock()
        mock_response.text = "content"
        mock_get.return_value = mock_response

        downloader = StructureDownloader(cache=None)

        with patch.object(downloader, "_url") as mock_url:
            mock_url.return_value = "http://example.com"
            downloader.download("  test  ", "cif")
            # Check that the stripped code was used
            mock_url.assert_called_with("test", "cif", "rcsb")

    @patch("requests.get")
    def test_format_with_dot_is_stripped(self, mock_get):
        mock_response = Mock()
        mock_response.text = "content"
        mock_get.return_value = mock_response

        downloader = StructureDownloader(cache=None)

        with patch.object(downloader, "_url") as mock_url:
            mock_url.return_value = "http://example.com"
            downloader.download("test", ".cif")
            # Check that the stripped format was used
            mock_url.assert_called_with("test", "cif", "rcsb")


# Test file writing to cache
class TestFileWriting:
    @patch("requests.get")
    def test_text_file_written_to_cache(self, mock_get, tmpdir):
        content = "test content"
        mock_response = Mock()
        mock_response.text = content
        mock_get.return_value = mock_response

        cache_dir = tmpdir.mkdir("cache")
        downloader = StructureDownloader(cache=str(cache_dir))

        result = downloader.download("test", "cif")

        assert isinstance(result, Path)
        assert result.read_text() == content

    @patch("requests.get")
    def test_binary_file_written_to_cache(self, mock_get, tmpdir):
        content = b"binary content"
        mock_response = Mock()
        mock_response.content = content
        mock_get.return_value = mock_response

        cache_dir = tmpdir.mkdir("cache")
        downloader = StructureDownloader(cache=str(cache_dir))

        result = downloader.download("test", "bcif")

        assert isinstance(result, Path)
        assert result.read_bytes() == content



import copy
import functools
import inspect
from abc import ABCMeta
from uuid import uuid1
import bpy
import databpy as db
import numpy as np
from databpy.object import LinkedObjectError
from mathutils import Vector
from ..blender import coll
from ..blender.utils import (
    get_viewport_region_from_context,
    new_bmesh,
    viewport_tag_redraw,
)
from ..nodes.material import append_material
from .base import BaseAnnotation
from .interface import AnnotationInterface
from .node_tree import annotations_node_tree
from .props import (
    BaseAnnotationProperties,
    create_annotation_type_inputs,
    create_property_interface,
)
from .utils import (
    get_all_class_annotations,
    get_blender_supported_type,
    get_view_matrix,
    is_perspective_projection,
)


def _validate_annotation_update(self, context, attr):
    """Update callback when annotation inputs change (both API and GUI)"""
    session = context.scene.MNSession
    entity = session.get(self.id_data.uuid)
    interface = entity.annotations._interfaces.get(self.uuid)
    instance = interface._instance
    # delete non blender attribute as blender attribute updated
    nbattr = f"_{attr}"
    if hasattr(instance, nbattr):
        delattr(instance, nbattr)
    try:
        if not instance.validate():
            raise ValueError(f"Invalid input {attr}")
    except Exception as exception:
        self.valid_inputs = False
        raise exception
    self.valid_inputs = True
    if instance._ready:
        # update annotation object
        entity.annotations._update_annotation_object()


class BaseAnnotationManager(metaclass=ABCMeta):
    """
    Base class for Annotation Manager

    This is the base class for the Annotation Manager, which manages all the
    annotation classes, instances and interfaces

    Entities that need annotation support have to derive from this base class and
    set the class attribute _entity_type to the entity tpye, the class attribute
    '_classes' and instance attribute '_interfaces' to empty dictionaries. Derived
    classes will have to pass the entity instance as part of its '__init__' as well.

    Attributes
    ----------
    visible: bool
        Whether to show or hide all annotations

    """

    _entity_type = None  # Derived classes need to specify
    _classes = {}  # All annotation classes across all entities

    def __init__(self, entity):
        # Access to the entity to which this manager is attached
        self._entity = entity
        self._draw_handler = None
        self._scene = None
        self._render_mode = False
        # annotation blender object
        self.bob = None
        # viewport params
        self._region = None
        self._rv3d = None
        # render params
        self._render_scale = 1.0
        self._image = None
        self._image_scale = 1
        # add draw handler by default
        self._draw_handler_add()

    def __iter__(self) -> iter:
        """To support iteration"""
        return iter(self._interfaces.values())

    def __len__(self) -> int:
        """Number of annotations"""
        return len(self._interfaces)

    def __getitem__(self, name: str) -> AnnotationInterface:
        """Return annotation interface based on name or index"""
        if isinstance(name, int):
            # numeric index based array access
            index = int(name)
            max = len(self._interfaces)
            if 0 <= index < max:
                key = list(self._interfaces.keys())[index]
                return self._interfaces[key]
            else:
                raise ValueError(f"Invalid index {index} of {max} items")
        return self.get(name)

    def _ipython_key_completions_(self) -> list[str]:
        """Return annotation names"""
        return [i.name for i in self._interfaces.values()]

    def __getstate__(self):
        """Get state for pickling this object"""
        state = self.__dict__.copy()
        # Remove reference to Scene before pickling
        del state["_scene"]
        return state

    def __setstate__(self, state):
        """Set state when unpickling this object"""
        self.__dict__.update(state)
        self._scene = None

    @classmethod
    def register(cls, annotation_class) -> None:
        """
        Register an annotation class

        This method adds the annotation class to the entity specific class
        registry (_classes) and adds a new method (add_<annotation_type>) to
        the manager with a signature that matches the annotation inputs.

        """
        cls._validate_annotation_class(annotation_class)
        # Add to Entity class specific registry
        annotation_type = annotation_class.annotation_type
        if annotation_type in cls._classes:
            raise ValueError(f"Annotation type {annotation_type} already registered")
        cls._classes[annotation_type] = annotation_class
        # Add to all annotation classes
        if cls._entity_type is None:
            raise ValueError("Entity type is needed in derived class")
        entity_annotation_type = f"{cls._entity_type.value}_{annotation_type}"
        if entity_annotation_type in BaseAnnotationManager._classes:
            raise ValueError(
                f"Annotation type {entity_annotation_type} already registered"
            )
        BaseAnnotationManager._classes[entity_annotation_type] = annotation_class
        # Add method to Entity specific manager
        method_name = f"add_{annotation_type}"

        # Add a dynamic wrapper to ensure custom signature is retained
        def dynamic_wrapper(cls, annotation_class, **kwargs):
            return cls._create_annotation_instance(annotation_class, **kwargs)

        method = functools.partialmethod(dynamic_wrapper, annotation_class)
        parameters = [
            inspect.Parameter("self", inspect.Parameter.POSITIONAL_ONLY),
            inspect.Parameter("annotation_class", inspect.Parameter.POSITIONAL_ONLY),
        ]
        # Add the annotation inputs as keyword params
        py_annotations = get_all_class_annotations(annotation_class)
        for attr, atype in py_annotations.items():
            default = inspect.Parameter.empty
            # set the default if the param in annotation class is a class
            # attribute and not just an annotation
            if hasattr(annotation_class, attr):
                default = getattr(annotation_class, attr)
            param = inspect.Parameter(
                attr, inspect.Parameter.KEYWORD_ONLY, annotation=atype, default=default
            )
            parameters.append(param)
        # Create method signature to match annotation class inputs
        method.func.__signature__ = inspect.Signature(
            parameters, return_annotation=AnnotationInterface
        )
        # Update annotation properties
        cls._update_annotation_props(annotation_class)
        setattr(cls, method_name, method)

    @classmethod
    def _update_annotation_props(cls, annotation_class: BaseAnnotation):
        """Update annotation properties attached to Object"""
        AnnotationProperties = type(
            "AnnotationProperties", (BaseAnnotationProperties,), {}
        )
        # Add each annotation type inputs as a pointer to a separate property group
        for annotation_type, annotation_class in BaseAnnotationManager._classes.items():
            update_callback = None
            # Add an update callback for annotation input properties to validate
            if hasattr(annotation_class, "validate") and callable(
                getattr(annotation_class, "validate")
            ):
                update_callback = _validate_annotation_update
            AnnotationInputs = create_annotation_type_inputs(
                annotation_class, update_callback=update_callback
            )
            bpy.utils.register_class(AnnotationInputs)
            AnnotationProperties.__annotations__[annotation_type] = (
                bpy.props.PointerProperty(type=AnnotationInputs)
            )
        # Re-register the new AnnotationProperties class
        bpy.utils.register_class(AnnotationProperties)
        # Re-assign the annotation properties to Object - old data is retained
        bpy.types.Object.mn_annotations = bpy.props.CollectionProperty(
            type=AnnotationProperties
        )

    @classmethod
    def unregister(cls, annotation_class) -> None:
        """
        Unregister a registered annotation class

        This method removes the annotation class from the entity speicific
        class registry and removes the 'add_<>' method from the manager

        """
        annotation_type = annotation_class.annotation_type
        if annotation_type not in cls._classes:
            raise ValueError(f"{annotation_class} is not registered")
        cls._validate_annotation_class(annotation_class)
        # Delete from Entity class specific registry
        del cls._classes[annotation_type]
        # Delete from all annotation classes
        entity_annotation_type = f"{cls._entity_type.value}_{annotation_type}"
        del BaseAnnotationManager._classes[entity_annotation_type]
        # Remove method from Entity specific manager
        method_name = f"add_{annotation_type}"
        delattr(cls, method_name)

    def get(self, name: str) -> AnnotationInterface:
        """Get an annotation by name"""
        for instance in self._interfaces.values():
            if instance.name == name:
                return instance
        raise ValueError(f"Annotation with name '{name}' not found")

    def remove(self, annotation: str | AnnotationInterface) -> None:
        """
        Remove an annotation by name or instance

        When a name is used, all annotations that match the name will be removed

        """
        instances = []
        if isinstance(annotation, str):
            for instance in self._interfaces.values():
                if instance.name == annotation:
                    instances.append(instance)
        else:
            if not isinstance(annotation, AnnotationInterface):
                raise ValueError("Invalid annotation instance")
            if annotation._uuid not in self._interfaces:
                raise ValueError(f"Instance for {annotation._uuid} not found")
            instances.append(annotation)
        if not instances:
            raise ValueError(f"No annotations found for {annotation}")
        for instance in instances:
            self._remove_annotation_instance(instance)

    def clear(self) -> None:
        """Remove all annotations"""
        instances = []
        for instance in self._interfaces.values():
            instances.append(instance)
        for instance in instances:
            self._remove_annotation_instance(instance)

    @property
    def visible(self) -> bool:
        """Visibility of all annotations - getter"""
        return self._entity.object.mn.annotations_visible

    @visible.setter
    def visible(self, value: bool) -> None:
        """Visibility of all annotations - setter"""
        self._entity.object.mn.annotations_visible = value
        if value:
            self._draw_handler_add()
        else:
            self._draw_handler_remove()
        self._update_annotation_object()
        viewport_tag_redraw()

    @classmethod
    def _validate_annotation_class(cls, annotation_class: BaseAnnotation) -> None:
        """Internal method to validate an annotation class"""
        if not issubclass(annotation_class, BaseAnnotation):
            raise ValueError(f"{annotation_class} is not an annotation class")
        if not hasattr(annotation_class, "annotation_type"):
            raise ValueError("annotation_type not specified in class")
        if annotation_class.draw.__qualname__ == "BaseAnnotation.draw":
            raise ValueError("draw method not found in class")

    def _create_annotation_instance(
        self, annotation_class: BaseAnnotation, **kwargs
    ) -> AnnotationInterface:
        """
        Create annotation instance and return a dynamic Annotation Interface

        This is the method that is bound to the 'add_<annotation_type>' methods
        of the manager. This method is responsible for creating the actual
        annotation instance, creating a Geometry Node with inputs that match the
        annotation type and creating an interface that binds the properties to
        the corresponding Annotation Geometry Node inputs.

        """
        # validations for required and invalid inputs
        py_annotations = get_all_class_annotations(annotation_class)
        for attr in py_annotations.keys():
            if not hasattr(annotation_class, attr) and attr not in kwargs:
                raise ValueError(f"{attr} is a required parameter")
        for attr in kwargs.keys():
            if attr not in py_annotations:
                raise ValueError(
                    f"Unknown input {attr}. Valid values are {py_annotations}"
                )
        # create an annotations instance
        annotation_instance = annotation_class(self._entity)
        # create a new dynamic interface class
        inteface_class_name = f"{annotation_class.__name__}_interface"
        DynamicInterface = type(inteface_class_name, (AnnotationInterface,), {})
        interface = DynamicInterface(annotation_instance)
        # set the interface attribute of the annotation instance so that
        # users can access the inputs and common params using this interface
        setattr(annotation_instance, "interface", interface)
        # instance is ready only after all the property interfaces are created
        setattr(annotation_instance, "_ready", False)
        # call the validate method in the annotation class if specified for
        # any annotation specific custom validation
        if hasattr(annotation_instance, "validate") and callable(
            getattr(annotation_instance, "validate")
        ):
            # set the annotation inputs to interface for validation
            for attr in py_annotations.keys():
                # set the input value based on what is passed
                # if input value is not passed, use the value from the annotation class
                value = kwargs.get(attr, None)
                if value is None:
                    value = getattr(annotation_class, attr, None)
                if value is not None:
                    setattr(interface.__class__, attr, value)
            # validate
            if not annotation_instance.validate():
                raise ValueError("Invalid annotation inputs")
        # only after all validations pass start doing real stuff like creating
        # properties and adding to the interface list
        uuid = str(uuid1())
        # add the new interface to the entity specific interfaces registry
        self._interfaces[uuid] = interface
        # create a new annotation property
        object = self._entity.object
        prop = object.mn_annotations.add()
        # use the instance uuid as name for later lookups using find()
        prop.name = uuid
        prop.type = annotation_class.annotation_type
        # set the annotations active index for GUI
        object.mn.annotations_active_index = len(object.mn_annotations) - 1
        # add the uuid as internal attribute
        # _uuid will be used to lookup this interface in the interfaces registry
        setattr(interface, "_uuid", uuid)
        # set the annotation name and create property interface
        value = kwargs.get("name", None)
        if value is not None:
            setattr(prop, "label", value)
        else:
            if object.mn.annotations_next_index:
                prop.label = f"Annotation.{object.mn.annotations_next_index:03d}"
            else:
                prop.label = "Annotation"
            object.mn.annotations_next_index += 1
        prop_interface = create_property_interface(self._entity, uuid, "label")
        setattr(interface.__class__, "name", prop_interface)
        # iterate though all the annotation inputs, set passed values and
        # create property interfaces
        entity_annotation_type = f"{self._entity_type.value}_{prop.type}"
        inputs = getattr(prop, entity_annotation_type, None)
        if inputs is not None:
            inputs.uuid = uuid  # add annotation uuid for lookup in update callback
            # link to the valid inputs property for use in draw handler
            prop_interface = create_property_interface(
                self._entity,
                uuid,
                "valid_inputs",
                annotation_type=entity_annotation_type,
            )
            setattr(interface.__class__, "_valid_inputs", prop_interface)
            # all annotation inputs as defined in class
            for attr, atype in py_annotations.items():
                # name is a special case that is already added above
                if attr == "name":
                    continue
                stype = get_blender_supported_type(atype)
                prop_interface = create_property_interface(
                    self._entity,
                    uuid,
                    attr,
                    stype,
                    annotation_instance,
                    entity_annotation_type,
                )
                setattr(interface.__class__, attr, prop_interface)
                # set the input value based on what is passed
                # if input value is not passed, use the value from the annotation class
                value = kwargs.get(attr, None)
                if value is None:
                    value = getattr(annotation_class, attr, None)
                if value is not None:
                    setattr(interface, attr, value)
        # create property interfaces for the common annotation params
        for item in prop.bl_rna.properties:
            if not item.is_runtime:
                continue
            prop_path = item.identifier
            if prop_path in ("name", "label", "type"):
                continue
            if item.type == "POINTER" and prop_path != "mesh_material":
                continue
            prop_interface = create_property_interface(self._entity, uuid, prop_path)
            setattr(interface.__class__, prop_path, prop_interface)
        # call the defaults method in the annotation class if specified
        if hasattr(annotation_instance, "defaults") and callable(
            getattr(annotation_instance, "defaults")
        ):
            annotation_instance.defaults()
        # material default needs to be set explicitly
        material = "MN Default"
        if material not in bpy.data.materials:
            append_material(material)
        interface.mesh_material = material
        # update annotation object
        self._update_annotation_object()
        # mark instance ready for object updates from property callbacks
        annotation_instance._ready = True
        # tag redraw of viewport to refresh values in GUI
        viewport_tag_redraw()
        # return the newly created dynamic annotation interface
        return interface

    def _remove_annotation_instance(self, instance) -> None:
        """Actual method to remove annotation instance"""
        uuid = instance._uuid
        object = self._entity.object
        index = object.mn_annotations.find(uuid)
        if index != -1:
            object.mn_annotations.remove(index)
        object.mn.annotations_active_index = len(object.mn_annotations) - 1
        # tag redraw of viewport to refresh values in GUI
        viewport_tag_redraw()
        # remove from instances registry
        del self._interfaces[uuid]
        # update annotation object
        self._update_annotation_object()

    def _remove_annotation_by_uuid(self, uuid: str) -> None:
        """Remove annotation instance by uuid - used by GUI"""
        if uuid not in self._interfaces:
            raise ValueError(f"Instance for {uuid} not found")
        self._remove_annotation_instance(self._interfaces[uuid])

    def _draw_handler_add(self):
        if bpy.app.background:
            return
        if self._draw_handler is not None:
            return
        self._draw_handler = bpy.types.SpaceView3D.draw_handler_add(
            self._draw_annotations_handler,
            (bpy.context,),
            "WINDOW",
            "POST_PIXEL",
        )
        self._redraw()

    def _draw_handler_remove(self):
        if self._draw_handler is not None:
            bpy.types.SpaceView3D.draw_handler_remove(self._draw_handler, "WINDOW")
            self._redraw()
            self._draw_handler = None

    def _redraw(self):
        if self._draw_handler is not None:
            viewport_tag_redraw()

    def _is_valid_entity(self) -> bool:
        try:
            _name = self._entity.name
            return True
        except LinkedObjectError:
            # remove any registered draw handler
            self._draw_handler_remove()
            return False

    def _draw_annotations_handler(self, context):
        if self._draw_handler is None:
            return
        if not self._is_valid_entity():
            return
        region, rv3d = get_viewport_region_from_context(context)
        # for viewport drawing, region and rv3d are required
        if region is None or rv3d is None:
            return
        self._region = region
        self._rv3d = rv3d
        self._scene = context.scene
        self._draw_annotations()

    def _enable_render_mode(self, scene, render_scale, image, image_scale):
        self._scene = scene
        self._render_scale = render_scale
        self._image = image
        self._image_scale = image_scale
        self._render_mode = True

    def _disable_render_mode(self):
        self._render_mode = False

    def _draw_annotations(self, get_geometry: bool = False):
        # all annotations visibilty
        if not self.visible:
            return
        # object visibility
        object = self._entity.object
        if object.hide_get():
            return
        # check if object is in current scene
        if self._scene is None:
            self._scene = bpy.context.scene
        if object.name not in self._scene.objects:
            return
        if get_geometry:
            # geometry of lines and bmesh objects
            geometry = {
                "lines": {
                    "vertices": [],
                    "edges": [],
                    "thickness": [],
                    "color": [],
                    "material_index": [],
                },
                "objects": {
                    "meshes": [],
                    "wireframe": [],
                    "thickness": [],
                    "color": [],
                    "material_index": [],
                    "shade_smooth": [],
                },
                "materials": {},
            }
            empty_geometry = copy.deepcopy(geometry)
        else:
            # calculate the min and max viewport distance of the bounding box verts
            # find bounding box verts
            bb_verts = [co for co in bpy.data.objects[object.name].bound_box]
            # viewport distance of the bounding box verts
            bb_verts_distances = []
            for vert in bb_verts:
                view_matrix = get_view_matrix(self)
                if is_perspective_projection(self):
                    # perspective
                    dist = (view_matrix @ Vector(vert)).length
                else:
                    # orthographic
                    dist = -(view_matrix @ Vector(vert)).z
                bb_verts_distances.append(dist)
            # min and max
            min_dist = min(bb_verts_distances)
            max_dist = max(bb_verts_distances)
        # iterate over all annotations
        for interface in self._interfaces.values():
            # annotation specific visibility
            if not interface.visible:
                continue
            # annotations input validity
            if not getattr(interface, "_valid_inputs", True):
                continue
            interface._instance.geometry = None
            if get_geometry:
                interface._instance.geometry = geometry
            else:
                # set distance params
                interface._instance._set_distance_params(min_dist, max_dist)
                if self._render_mode:
                    # set the render params
                    interface._instance._set_render_params(
                        self._scene, self._render_scale, self._image, self._image_scale
                    )
                else:
                    # set the viewport region params
                    interface._instance._set_viewport_params(
                        self._scene, self._region, self._rv3d
                    )
            # handle exceptions to allow other annotations to be drawn
            try:
                interface._instance.draw()
            except Exception:
                pass
        # check and return geometry if present
        if get_geometry and geometry != empty_geometry:
            return geometry

    def _annotation_object_exists(self) -> bool:
        """Internal: Check if annotation object exists"""
        exists = True
        if self.bob is not None:
            try:
                _ = self.bob.object.name
            except LinkedObjectError:
                exists = False
        else:
            exists = False
        return exists

    def _create_annotation_object(self):
        """Internal: Create annotation object if it doesn't exist"""
        name = f"MN_an_{self._entity.object.name}"
        # create an empty object
        object = db.create_object(name=name, collection=coll.mn())
        # create geometry nodes modifier
        object.modifiers.new("MN_Annotations", "NODES")
        # attach the annotations node tree
        tree = annotations_node_tree()
        tree.name = name
        object.modifiers[0].node_group = tree
        object.parent = self._entity.object
        # save the BlenderObject for further updates
        self.bob = db.BlenderObject(object)

    def _clear_annotation_object(self):
        """Internal: Clear annotation object"""
        if len(self.bob.object.data.vertices) == 0:
            return
        # clear existing geometry
        self.bob.object.data.clear_geometry()
        # clear existing material slots
        self.bob.object.data.materials.clear()

    def _update_annotation_object(self):
        """Internal: Update the annotation object mesh and attributes"""
        # get the geometry from all annotations
        geometry = self._draw_annotations(get_geometry=True)

        # check and clear or create annotation object
        if self._annotation_object_exists():
            self._clear_annotation_object()
            if geometry is None:
                return
        else:
            if geometry is None:
                return
            self._create_annotation_object()

        # create a new bmesh
        with new_bmesh() as bm:
            # convenience methods
            add_vert = bm.verts.new
            add_edge = bm.edges.new
            add_face = bm.faces.new

            # edge attributes
            attr_is_line = []
            attr_color = []
            attr_thickness = []
            attr_material_index = []
            attr_shade_smooth = []

            # add lines
            lines = geometry["lines"]
            for v in lines["vertices"]:
                add_vert(v)
            bm.verts.ensure_lookup_table()
            for i, (i1, i2) in enumerate(lines["edges"]):
                add_edge((bm.verts[i1], bm.verts[i2]))
                # add edge attributes
                attr_is_line.append(True)
                attr_color.append(lines["color"][i])
                attr_thickness.append(lines["thickness"][i])
                attr_material_index.append(lines["material_index"][i])
                attr_shade_smooth.append(True)

            # add bmesh objects
            # From: https://blender.stackexchange.com/questions/50160/scripting-low-level-join-meshes-elements-hopefully-with-bmesh
            objects = geometry["objects"]
            for i, bm_to_add in enumerate(objects["meshes"]):
                try:
                    offset = len(bm.verts)
                    # add verts
                    for v in bm_to_add.verts:
                        try:
                            add_vert(v.co)
                        except ValueError:
                            # vertex exists!
                            pass
                    bm.verts.index_update()
                    bm.verts.ensure_lookup_table()
                    object_wireframe = objects["wireframe"][i]
                    # add faces
                    if not object_wireframe and bm_to_add.faces:
                        for face in bm_to_add.faces:
                            try:
                                add_face(
                                    (bm.verts[i.index + offset] for i in face.verts)
                                )
                            except ValueError:
                                # face exists! - can happen with n=2 (flat) objects
                                pass
                        bm.faces.index_update()
                    # add edges
                    if bm_to_add.edges:
                        object_color = objects["color"][i]
                        object_material_index = objects["material_index"][i]
                        object_shade_smooth = objects["shade_smooth"][i]
                        for edge in bm_to_add.edges:
                            edge_seq = (bm.verts[i.index + offset] for i in edge.verts)
                            try:
                                add_edge(edge_seq)
                            except ValueError:
                                # edge exists!
                                pass
                            # add edge attributes
                            if objects["wireframe"][i]:
                                attr_is_line.append(True)
                                attr_thickness.append(objects["thickness"][i])
                            else:
                                attr_is_line.append(False)
                                attr_thickness.append(0.0)
                            attr_color.append(object_color)
                            attr_material_index.append(object_material_index)
                            attr_shade_smooth.append(object_shade_smooth)
                        bm.edges.index_update()
                finally:
                    # free bmesh
                    bm_to_add.free()

            # update the object mesh
            bm.to_mesh(self.bob.object.data)

        # add materials to object material slots
        for material_name in geometry["materials"]:
            material = bpy.data.materials[material_name]
            # add material slot
            self.bob.object.data.materials.append(material)

        # add edge attributes
        if len(attr_is_line) > 0:
            # add the is_line attribute to distinguish lines
            self.bob.store_named_attribute(
                np.array(attr_is_line),
                name="is_line",
                atype=db.AttributeTypes.BOOLEAN,
                domain=db.AttributeDomains.EDGE,
            )
            # add the Color attribute for line color
            self.bob.store_named_attribute(
                np.array(attr_color),
                name="Color",
                atype=db.AttributeTypes.FLOAT_COLOR,
                domain=db.AttributeDomains.EDGE,
            )
            # add a thickness attribute for line thickness
            self.bob.store_named_attribute(
                np.array(attr_thickness),
                name="thickness",
                atype=db.AttributeTypes.FLOAT,
                domain=db.AttributeDomains.EDGE,
            )
            # add material_slot_index attribute
            # Note: material_index is a reserved attribute
            self.bob.store_named_attribute(
                np.array(attr_material_index),
                name="material_slot_index",
                atype=db.AttributeTypes.INT,
                domain=db.AttributeDomains.EDGE,
            )
            # add the shade_smooth attribute
            self.bob.store_named_attribute(
                np.array(attr_shade_smooth),
                name="shade_smooth",
                atype=db.AttributeTypes.BOOLEAN,
                domain=db.AttributeDomains.EDGE,
            )



"""Trajectory entity for molecular dynamics simulations.

Provides Trajectory class for loading and visualizing MD trajectories in Blender
using MDAnalysis.
"""

import functools
import logging
from pathlib import Path
from typing import Callable, Dict
import bpy
import databpy as db
import MDAnalysis as mda
import numpy as np
from MDAnalysis.core.groups import AtomGroup
from ...assets import data
from ...blender import coll, path_resolve, set_obj_active
from ...blender import utils as blender_utils
from ...nodes.geometry import (
    add_style_branch,
)
from ...nodes.nodes import styles_mapping
from ...nodes.styles import (
    StyleBase,
)
from ...utils import (
    temp_override_property,
)
from ..base import EntityType, MolecularEntity
from ..utilities import (
    BoolObjectMNProperty,
    IntObjectMNProperty,
    StringObjectMNProperty,
    _unique_aname,
    _validate_non_negative,
)
from .annotations import TrajectoryAnnotationManager
from .helpers import FrameManager, _ag_to_bool
from .selections import SelectionManager

logger = logging.getLogger(__name__)


class Trajectory(MolecularEntity):
    """MD trajectory entity for Blender visualization.

    Complete interface for loading, visualizing, and manipulating MD trajectories
    in Blender using MDAnalysis.

    Features: trajectory loading, attribute computation, selection management,
    visual styling, frame interpolation/averaging, periodic boundary handling,
    Blender animation integration.

    Attributes
    ----------
    universe : mda.Universe
        MDAnalysis Universe with topology and trajectory
    frame_manager : FrameManager
        Position caching and frame updates
    selections : SelectionManager
        Dynamic atom selections
    calculations : dict
        Custom per-frame calculations
    annotations : TrajectoryAnnotationManager
        Trajectory annotations
    world_scale : float
        Scale factor from Angstroms to Blender units
    frame : int
        Current animation frame (synced with Blender)
    subframes : int
        Interpolation steps between frames
    offset : int
        Frame offset for playback
    average : int
        Number of frames to average (smoothing)
    correct_periodic : bool
        Apply periodic boundary corrections
    interpolate : bool
        Enable position interpolation

    Examples
    --------
    ```{python}
    #| warning: false
    import MDAnalysis as mda
    from MDAnalysis.tests.datafiles import PSF, DCD
    import molecularnodes as mn
    canvas = mn.Canvas()
    u = mda.Universe(PSF, DCD)
    traj = mn.entities.Trajectory(u)
    traj.add_style(mn.StyleSpheres(geometry="Mesh"), selection="resname LYS")
    canvas.frame_view(traj)
    canvas.snapshot()
    ```
    """

    # Blender property descriptors with validation
    frame = IntObjectMNProperty("frame")
    subframes = IntObjectMNProperty("subframes", validate_fn=_validate_non_negative)
    offset = IntObjectMNProperty("offset")
    average = IntObjectMNProperty("average", validate_fn=_validate_non_negative)
    correct_periodic = BoolObjectMNProperty("correct_periodic")
    interpolate = BoolObjectMNProperty("interpolate")

    _mn_frame = BoolObjectMNProperty("frame_hidden")
    _mn_styles_active_index = IntObjectMNProperty(
        "styles_active_index", _validate_non_negative
    )
    _mn_entity_type = StringObjectMNProperty("entity_type")
    _mn_filepath_topology = StringObjectMNProperty("filepath_topology")
    _mn_filepath_trajectory = StringObjectMNProperty("filepath_trajectory")
    _mn_n_frames = IntObjectMNProperty("n_frames", _validate_non_negative)

    def __init__(
        self,
        universe: mda.Universe,
        name: str = "NewUniverseObject",
        world_scale: float = 0.01,
        create_object: bool = True,
    ):
        """Initialize Trajectory from MDAnalysis Universe.

        Parameters
        ----------
        universe : mda.Universe
            MDAnalysis Universe with topology and trajectory
        name : str, default="NewUniverseObject"
            Name for the Blender object
        world_scale : float, default=0.01
            Scale factor from Angstroms to Blender units
        create_object : bool, default=True
            Whether to immediately create the Blender object

        Notes
        -----
        Default world_scale of 0.01 converts Angstroms to Blender units.
        """
        super().__init__()
        self.universe: mda.Universe = universe
        self.selections: SelectionManager = SelectionManager(self)
        self.calculations: Dict[str, Callable] = {}
        self.world_scale = world_scale
        self._updating_in_progress = False
        self.annotations = TrajectoryAnnotationManager(self)
        self.frame_manager = FrameManager(self)
        if create_object:
            self.create_object(name=name)

    @property
    def _is_orthorhombic(self) -> bool:
        """Check if simulation box is orthorhombic.

        Orthorhombic boxes (all angles = 90°) enable optimized periodic boundary handling.

        Returns
        -------
        bool
            True if box angles are all 90 degrees
        """
        dim = self.universe.dimensions
        if dim is None:
            return False
        return np.allclose(dim[3:], 90.0)

    @property
    def atoms(self) -> mda.AtomGroup:
        """All atoms as MDAnalysis AtomGroup."""
        if self.universe.atoms is None:
            raise ValueError(f"Universe {self.universe} has no atoms")
        return self.universe.atoms

    @property
    def _scaled_position(self) -> np.ndarray:
        """Current atom positions in Blender world units.

        Returns
        -------
        np.ndarray
            Shape (n_atoms, 3) in Blender units
        """
        return self.atoms.positions * self.world_scale

    @property
    def uframe(self) -> int:
        """Current frame number in MDAnalysis Universe.

        Differs from `frame` property (Blender scene frame). Query actual trajectory frame.

        Returns
        -------
        int
            Current Universe frame (0-indexed)
        """
        return self.universe.trajectory.frame

    @uframe.setter
    def uframe(self, value: int) -> None:
        """Set current frame in MDAnalysis Universe.

        Only updates if changed to avoid redundant operations.

        Parameters
        ----------
        value : int
            Target frame number (0-indexed)
        """
        if self.universe.trajectory.frame != value:
            self.universe.trajectory[value]

    @functools.cached_property
    def _elements(self) -> np.ndarray:
        """
        Cached computation of element symbols for all atoms.
        This is computed once and reused by multiple attribute computations.
        """
        if hasattr(self.atoms, "elements"):
            return self.atoms.elements

        try:
            default_guesser = mda.guesser.default_guesser.DefaultGuesser(None)  # type: ignore
            guessed_elements = [
                x
                if x in data.elements.keys()
                else default_guesser.guess_atom_element(x)
                for x in self.atoms.names  # type: ignore
            ]
            return np.array(guessed_elements)
        except Exception as e:
            logger.warning(f"Failed to compute elements, using placeholder 'X': {e}")
            return np.repeat("X", len(self))

    def _compute_elements(self) -> np.ndarray:
        """Return cached elements (for backwards compatibility)"""
        return self._elements

    def _compute_atomic_number(self) -> np.ndarray:
        return np.array(
            [
                data.elements.get(element, data.elements["X"]).get("atomic_number")
                for element in self._elements
            ]
        )

    def _compute_vdw_radii(self) -> np.ndarray:
        return (
            np.array(
                [
                    data.elements.get(element, {}).get("vdw_radii", 100)
                    for element in self._elements
                ]
            )
            * 0.01  # pm to Angstrom
            * self.world_scale  # Angstrom to world scale
        )

    def _compute_mass(self) -> np.ndarray:
        # units: daltons
        if hasattr(self.atoms, "masses"):
            return np.array([x.mass for x in self.atoms])
        else:
            masses = [
                data.elements.get(element, {"standard_mass": 0}).get("standard_mass")
                for element in self._elements
            ]
            return np.array(masses)

    def _compute_res_name(self) -> np.ndarray:
        return np.array(list(map(lambda x: x[0:3], self.atoms.resnames)))

    def _compute_res_name_int(self) -> np.ndarray:
        res_name = self._compute_res_name()
        return np.array(
            [
                data.residues.get(name, data.residues["UNK"]).get("res_name_num")
                for name in res_name
            ],
            dtype=int,
        )

    def _compute_b_factor(self) -> np.ndarray:
        if hasattr(self.atoms, "tempfactors"):
            return self.atoms.tempfactors
        else:
            return np.zeros(len(self))

    def _compute_occupancy(self) -> np.ndarray:
        # corresponds to Occupancies topology attr
        if hasattr(self.atoms, "occupancies"):
            return self.atoms.occupancies
        else:
            return np.zeros(len(self))

    def _compute_charge(self) -> np.ndarray:
        # corresponds to Charges topology attr
        if hasattr(self.atoms, "charges"):
            return self.atoms.charges
        else:
            return np.zeros(len(self))

    def _compute_res_id(self) -> np.ndarray:
        return self.atoms.resids

    def _compute_atom_id(self) -> np.ndarray:
        return self.atoms.ids

    def _compute_segindices(self) -> np.ndarray:
        segs = []
        for seg in self.atoms.segments:
            segs.append(seg.atoms[0].segid)

        else:
            try:
                self.object["segments"] = segs
            except db.LinkedObjectError:
                logger.warning("Failed to store segments metadata on object")

        return self.atoms.segindices

    def _compute_chain_id_int(self) -> np.ndarray:
        chain_ids, chain_id_index = np.unique(self.atoms.chainIDs, return_inverse=True)

        try:
            self.object["chain_ids"] = chain_ids.astype(str).tolist()
        except db.LinkedObjectError:
            logger.warning("Failed to store chain_ids metadata on object")

        return chain_id_index

    def _compute_atom_type_int(self) -> np.ndarray:
        atom_type_unique, atom_type_index = np.unique(
            self.atoms.types, return_inverse=True
        )

        try:
            self.object["atom_type_unique"] = atom_type_unique
        except db.LinkedObjectError:
            logger.warning("Failed to store atom_type_unique metadata on object")

        return atom_type_index

    def _compute_atom_name_int(self) -> np.ndarray:
        if hasattr(self.atoms, "names"):
            return np.array(
                [data.atom_names.get(x, -1) for x in self.atoms.names],  # type: ignore
                dtype=int,
            )
        else:
            return np.repeat(int(-1), len(self))

    def _compute_is_lipid(self) -> np.ndarray:
        return np.isin(self.atoms.resnames, data.lipid_names)

    def _save_filepaths_on_object(self) -> None:
        """Save file paths to the Blender object for reference"""
        if isinstance(self.universe.filename, (str, Path)):
            self._mn_filepath_topology = str(path_resolve(self.universe.filename))
        if isinstance(self.universe.trajectory.filename, (str, Path)):
            self._mn_filepath_trajectory = str(
                path_resolve(self.universe.trajectory.filename)
            )

    def reset_playback(self) -> None:
        """Set the playback settings to their default values"""
        self.subframes = 0
        self.offset = 0
        self.average = 0
        self.correct_periodic = False
        self.interpolate = False

    @property
    def _blender_attributes(self) -> Dict[str, Callable | str]:
        """Registry of default attributes for Blender object.

        Defines standard molecular attributes computed at trajectory creation.
        Maps attribute names to compute functions or MDAnalysis selection strings.

        Returns
        -------
        dict
            Attribute names to compute functions or selection strings
        """
        return {
            "atomic_number": self._compute_atomic_number,
            "vdw_radii": self._compute_vdw_radii,
            "mass": self._compute_mass,
            "res_id": self._compute_res_id,
            "segid": self._compute_segindices,
            "res_name": self._compute_res_name_int,
            "atom_id": self._compute_atom_id,
            "b_factor": self._compute_b_factor,
            "occupancy": self._compute_occupancy,
            "charge": self._compute_charge,
            "chain_id": self._compute_chain_id_int,
            "atom_types": self._compute_atom_type_int,
            "atom_name": self._compute_atom_name_int,
            "is_backbone": "backbone or nucleicbackbone or name BB",
            "is_alpha_carbon": "name CA or name BB",
            "is_solvent": "name OW or name HW1 or name HW2 or resname W or resname PW",
            "is_nucleic": "nucleic",
            "is_lipid": self._compute_is_lipid,
            "is_peptide": "protein or (name BB SC*)",
        }

    def _store_default_attributes(self) -> None:
        """Store default attributes"""

        for name, item in self._blender_attributes.items():
            try:
                if isinstance(item, str):
                    data = _ag_to_bool(self.universe.select_atoms(item))
                elif callable(item):
                    data = item()
                else:
                    raise ValueError("Unable to convert to attribute for storage")
                self.store_named_attribute(
                    data=data,
                    name=name,
                )
            except (mda.NoDataError, AttributeError) as e:
                logger.debug(f"Skipping attribute '{name}': {e}")
            except Exception as e:
                logger.warning(f"Failed to compute attribute '{name}': {e}")

    def _store_extra_attributes(self) -> None:
        # TODO: enable adding of arbitrary mda.Universe attirbutes not currently applied
        pass

    def _create_object(
        self,
        name: str = "NewUniverseObject",
    ) -> None:
        """Create Blender mesh object (internal).

        Creates mesh with positions and bonds, stores attributes, sets up modifiers.

        Parameters
        ----------
        name : str
            Name for the Blender object
        """
        self.object = db.create_object(
            name=name,
            collection=coll.mn(),
            vertices=self._scaled_position,
            edges=self.atoms.bonds.indices if hasattr(self.atoms, "bonds") else None,
        )
        self._mn_entity_type = EntityType.MD.value

        self._store_default_attributes()
        self._store_extra_attributes()
        self._setup_modifiers()

    def create_object(self, name: str = "NewUniverseObject") -> bpy.types.Object:
        """Create and initialize Blender object for trajectory.

        Creates mesh, computes attributes, sets up modifiers, registers with MolecularNodes.

        Parameters
        ----------
        name : str, default="NewUniverseObject"
            Name for the Blender object

        Returns
        -------
        bpy.types.Object
            Created Blender object
        """
        self._create_object(name=name)

        self._mn_n_frames = self.universe.trajectory.n_frames
        self._save_filepaths_on_object()
        set_obj_active(self.object)

        return self.object

    def _update_calculations(self) -> None:
        """Update all registered calculations for the current frame"""
        for name, func in self.calculations.items():
            try:
                self.store_named_attribute(data=func(self.universe), name=name)
            except Exception as e:
                logger.error(
                    f"Failed to update calculation '{name}': {e}", exc_info=True
                )

    def _update_selections(self) -> None:
        """Update all selections for the current frame."""
        for item in self.selections.items:
            try:
                # Skip non-updating selections - they use static masks
                if not item.updating:
                    continue

                # Lazy initialization will occur if needed
                selection = self.selections.get(item.name)
                if selection is None:
                    raise KeyError(f"Selection '{item.name}' not found")

                # Don't recreate atomgroup for immutable selections (created from atomgroups)
                # These selections maintain their own atomgroup reference
                if not item.immutable:
                    selection.set_atom_group(item.string)
                selection.set_selection()
            except KeyError as e:
                logger.warning(
                    f"Failed to update selection '{item.name}': {e}. Skipping this selection."
                )
            except mda.SelectionError as e:
                logger.error(
                    f"Invalid selection syntax for '{item.name}': {e}. Skipping this selection."
                )
            except Exception as e:
                logger.error(
                    f"Error updating selection '{item.name}': {e}. Skipping this selection.",
                    exc_info=True,
                )

    def set_frame(self, frame: int) -> None:
        """Update trajectory state for scene frame.

        Main entry point called by Blender's animation system. Updates positions,
        selections, and calculations with recursion prevention.

        Parameters
        ----------
        frame : int
            Scene frame number (mapping applied to get Universe frame)

        Notes
        -----
        Typically called automatically by frame change handlers, not user code.
        """
        if self._updating_in_progress:
            logger.debug("Update already in progress, skipping nested update")
            return

        try:
            self._updating_in_progress = True
            self._update_positions(frame)
            self._update_selections()
            self._update_calculations()
            # update annotation object
            self.annotations._update_annotation_object()
        finally:
            self._updating_in_progress = False

    def _update_positions(self, frame: int) -> None:
        """
        Internal method to update atom positions.

        Delegates to FrameManager which handles caching, interpolation,
        averaging, and periodic corrections.

        Args:
            frame: Scene frame number
        """
        self.position = self.frame_manager.get_positions_at_frame(frame)

    def __repr__(self) -> str:
        return f"<Trajectory, `universe`: {self.universe}, `object`: {self.object}"

    def add_style(
        self,
        style: StyleBase | str = "spheres",
        color: str | None = "common",
        selection: str | AtomGroup | None = None,
        material: bpy.types.Material | str | None = None,
        name: str | None = None,
    ) -> "Trajectory":
        """
        Add a visual style to the trajectory.

        Parameters
        ----------
        style : bpy.types.GeometryNodeTree | str, optional
            The style to apply to the trajectory. Can be a GeometryNodeTree or a string
            identifying a predefined style (e.g., "spheres", "sticks", "ball_stick").
            Default is "spheres".

        color : str | None, optional
            The coloring scheme to apply. Can be "common" (element-based coloring),
            "chain", "residue", or other supported schemes. If None, no coloring
            is applied. Default is "common".

        selection : str | AtomGroup | None, optional
            Apply the style only to atoms matching this selection. Can be:
            - A string referring to an existing boolean attribute on the trajectory
            - A AtomGroup object defining a selection criteria
            - None to apply to all atoms (default)

        material : bpy.types.Material | str | None, optional
            The material to apply to the styled atoms. Can be a Blender Material object,
            a string with a material name, or None to use default materials. Default is None.

        name: str, optional
            The label for this style

        Returns
        -------
        Trajectory
            Returns self for method chaining.

        Raises
        ------
        ValueError
            If an unsupported style string is passed

        Notes
        -----
        If a selection is provided, it will be evaluated and stored as a new
        named attribute on the trajectory with an automatically generated name (sel_N).
        """
        if style is None:
            return self

        if isinstance(style, str) and style not in styles_mapping:
            raise ValueError(
                f"Invalid style '{style}'. Supported styles are {[key for key in styles_mapping.keys()]}"
            )
        if selection is None:
            sel_name = None
        else:
            att_name = _unique_aname(self.object, "sel")
            if isinstance(selection, str):
                # TODO: There are currently no validations for the selection phrase
                sel_name = self.selections.add(selection, att_name).name
            elif isinstance(selection, AtomGroup):
                sel_name = self.selections.from_atomgroup(selection, name=att_name).name
            # TODO: Delete these named attributes when style is deleted
            # Currently, styles are removed using GeometryNodeInterFace.remove(),

        node_style = add_style_branch(
            tree=self.tree,
            style=style,
            color=color,
            selection=sel_name,
            material=material,
            name=name,
        )

        # set the active index for UI to the newly added style
        self._mn_styles_active_index = self.tree.nodes.find(node_style.name)

        return self

    def _get_3d_bbox(self, selection: mda.AtomGroup | None) -> list[tuple]:
        """Get the 3D bounding box vertices of atoms in an AtomGroup"""
        if selection is None:
            return blender_utils.get_bounding_box(self.object)
        v0, v1 = selection.bbox() * self.world_scale
        bb_verts_3d = [
            (v0[0], v0[1], v0[2]),
            (v0[0], v0[1], v1[2]),
            (v0[0], v1[1], v0[2]),
            (v0[0], v1[1], v1[2]),
            (v1[0], v0[1], v0[2]),
            (v1[0], v0[1], v1[2]),
            (v1[0], v1[1], v0[2]),
            (v1[0], v1[1], v1[2]),
        ]
        return bb_verts_3d

    def get_view(
        self, selection: str | AtomGroup | None = None, frame: int | None = None
    ) -> list[tuple]:
        """
        Get the 3D bounding box of a selection within the trajectory

        Parameters
        ----------

        selection : str | AtomGroup, optional
            A selection phrase or AtomGroup
            When not specified, the whole entity is considered

        frame: int, optional
            Frame number of trajectory to use for calculating bounds.
            When not specified, current trajectory frame is used

        """
        if frame is not None:
            if frame < 0 or frame >= self.universe.trajectory.n_frames:
                raise ValueError(
                    f"{frame} is not within range [0, {self.universe.trajectory.n_frames - 1}]"
                )
        else:
            frame = self.uframe
        # temporarily set trajectory frame to specified value
        with temp_override_property(self, "uframe", frame):
            if selection is None:
                # return bbox of object when no selection specified
                return self._get_3d_bbox(None)
            if isinstance(selection, AtomGroup):
                atom_group = selection
            elif isinstance(selection, str):
                # allow multiple comma separated selection phrases as well
                selection_array = selection.split(",")
                try:
                    atom_group = self.universe.select_atoms(*selection_array)
                except Exception:
                    raise ValueError(f"Invalid {selection} phrase")
            else:
                raise ValueError(f"{selection} is neither a str or AtomGroup")

            if atom_group.n_atoms == 0:
                # return bbox of object when selection is empty
                return self._get_3d_bbox(None)

            # return the 3D bounding box vertices of the selected AtomGroup
            return self._get_3d_bbox(atom_group)



import biotite.structure as struc
import numpy as np
from biotite import InvalidFileError
from biotite.structure import (
    BadStructureError,
    annotate_sse,
    connect_via_residue_names,
    spread_residue_wise,
)
from biotite.structure.io import pdb
from .assembly import AssemblyParser
from .reader import ReaderBase


class PDBReader(ReaderBase):
    extra_fields = ["b_factor", "occupancy", "charge", "atom_id"]
    _extra_annotations = {}

    @staticmethod
    def set_extra_annotations(annotations, array, file):
        return array

    def read(self, file_path):
        return pdb.PDBFile.read(file_path)

    def get_structure(self, model: int | None = None):
        # a bit dirty, but we first try and get the bond information from the file
        # if that fails, then we extract without the bonds and try to create bonds based
        # on residue / atom names.
        try:
            array = pdb.get_structure(
                pdb_file=self.file,
                model=model,
                extra_fields=self.extra_fields,
                include_bonds=True,
            )
        except AttributeError as e:
            print(
                f"Unable to get bond information: {e}\nAttempting `connect_via_residue_names()`"
            )
            array = pdb.get_structure(
                pdb_file=self.file,
                model=model,
                extra_fields=self.extra_fields,
                include_bonds=False,
            )
            try:
                array.bonds = connect_via_residue_names(array)
            except AttributeError as e:
                print("Not able to find bonds via residue: {e}")

        try:
            sec_struct = _get_sec_struct(self.file, array)
        except BadStructureError:
            sec_struct = _comp_secondary_structure(array[0])

        array.set_annotation("sec_struct", sec_struct)

        if array is None:
            raise InvalidFileError("Unable to read PDB file.")

        return array

    def _assemblies(self):
        return PDBAssemblyParser(self.file).get_assemblies()


def _get_sec_struct(file, array):
    lines = np.array(file.lines)
    lines_helix = lines[np.char.startswith(lines, "HELIX")]
    lines_sheet = lines[np.char.startswith(lines, "SHEET")]
    if len(lines_helix) == 0 and len(lines_sheet) == 0:
        raise struc.BadStructureError("No secondary structure information detected.")

    sec_struct = np.zeros(array.array_length(), int)

    helix_values = (22, 25, 34, 37, 20)
    sheet_values = (23, 26, 34, 37, 22)

    values = ((lines_helix, 1, helix_values), (lines_sheet, 2, sheet_values))

    def _get_mask(line, start1, end1, start2, end2, chainid):
        """
        Takes a line and makes a mask for the array for atoms which are defined inside the
        range of values defined by the start and end values.
        """
        # bump the starting values down by one for indexing into
        # pythong strings
        start1 -= 1
        start2 -= 1
        chainid -= 1

        # get the values as intergers
        start_num = int(line[start1:end1])
        end_num = int(line[start2:end2])
        chain_id = line[chainid].strip()

        # create a mask for the array based on these values
        mask = np.logical_and(
            np.logical_and(array.chain_id == chain_id, array.res_id >= start_num),
            array.res_id <= end_num,
        )

        return mask

    # assigns secondary structure based on lines and stores it in the sec_struct
    for lines, idx, value_list in values:
        for line in lines:
            sec_struct[_get_mask(line, *value_list)] = idx

    # assign remaining AA atoms to 3 (loop), while all other remaining
    # atoms will be 0 (not relevant)
    mask = np.logical_and(sec_struct == 0, struc.filter_canonical_amino_acids(array))

    sec_struct[mask] = 3

    return sec_struct


def _comp_secondary_structure(array):
    """Use dihedrals to compute the secondary structure of proteins

    Through biotite built-in method derivated from P-SEA algorithm (Labesse 1997)
    Returns an array with secondary structure for each atoms where:
    - 0 = '' = non-protein or not assigned by biotite annotate_sse
    - 1 = a = alpha helix
    - 2 = b = beta sheet
    - 3 = c = coil

    Inspired from https://www.biotite-python.org/examples/gallery/structure/transketolase_sse.html
    """
    # TODO Port [PyDSSP](https://github.com/ShintaroMinami/PyDSSP)

    conv_sse_char_int = {"a": 1, "b": 2, "c": 3, "": 0}

    char_sse = annotate_sse(array)
    int_sse = np.array([conv_sse_char_int[char] for char in char_sse], dtype=int)
    atom_sse = spread_residue_wise(array, int_sse)

    return atom_sse


class PDBAssemblyParser(AssemblyParser):
    # Implementation adapted from ``biotite.structure.io.pdb.file``

    def __init__(self, pdb_file):
        self._file = pdb_file

    def list_assemblies(self):
        return self._file.list_assemblies()

    def get_transformations(self, assembly_id):
        # Get lines containing transformations for assemblies
        remark_lines = self._file.get_remark(350)
        if remark_lines is None:
            raise InvalidFileError(
                "File does not contain assembly information (REMARK 350)"
            )
        # Get lines corresponding to selected assembly ID
        assembly_start_i = None
        assembly_stop_i = None
        for i, line in enumerate(remark_lines):
            if line.startswith("BIOMOLECULE"):
                current_assembly_id = line[12:].strip()
                if assembly_start_i is not None:
                    # Start was already found -> this is the next entry
                    # -> this is the stop
                    assembly_stop_i = i
                    break
                if current_assembly_id == assembly_id:
                    assembly_start_i = i
        # In case of the final assembly of the file,
        # the 'stop' is the end of REMARK 350 lines
        assembly_stop_i = len(remark_lines) if assembly_stop_i is None else i
        if assembly_start_i is None:
            raise KeyError(f"The assembly ID '{assembly_id}' is not found")
        assembly_lines = remark_lines[assembly_start_i:assembly_stop_i]

        # Get transformations for a sets of chains
        transformations = []
        chain_set_start_indices = [
            i
            for i, line in enumerate(assembly_lines)
            if line.startswith("APPLY THE FOLLOWING TO CHAINS")
        ]
        # Add exclusive stop at end of records
        chain_set_start_indices.append(len(assembly_lines))
        for i in range(len(chain_set_start_indices) - 1):
            start = chain_set_start_indices[i]
            stop = chain_set_start_indices[i + 1]
            # Read affected chain IDs from the following line(s)
            affected_chain_ids = []
            transform_start = None
            for j, line in enumerate(assembly_lines[start:stop]):
                if line.startswith("APPLY THE FOLLOWING TO CHAINS:") or line.startswith(
                    "                   AND CHAINS:"
                ):
                    affected_chain_ids += [
                        chain_id.strip() for chain_id in line[30:].split(",")
                    ]
                else:
                    # Chain specification has finished
                    # BIOMT lines start directly after chain specification
                    transform_start = start + j
                    break
            # Parse transformations from BIOMT lines
            if transform_start is None:
                raise InvalidFileError("No 'BIOMT' records found for chosen assembly")

            matrices = _parse_transformations(assembly_lines[transform_start:stop])

            for matrix in matrices:
                transformations.append((affected_chain_ids, matrix.tolist()))

        return transformations

    def get_assemblies(self):
        assembly_dict = {}
        for assembly_id in self.list_assemblies():
            assembly_dict[assembly_id] = self.get_transformations(assembly_id)

        return assembly_dict


def _parse_transformations(lines):
    """
    Parse the rotation and translation transformations from
    *REMARK* 290 or 350.
    Return as array of matrices and vectors respectively
    """

    # Each transformation requires 3 lines for the (x,y,z) components
    if len(lines) % 3 != 0:
        raise InvalidFileError("Invalid number of transformation vectors")
    n_transformations = len(lines) // 3

    matrices = np.tile(np.identity(4), (n_transformations, 1, 1))

    transformation_i = 0
    component_i = 0
    for line in lines:
        # The first two elements (component and
        # transformation index) are not used
        transformations = [float(e) for e in line.split()[2:]]

        if len(transformations) != 4:
            raise InvalidFileError("Invalid number of transformation vector elements")
        matrices[transformation_i, component_i, :] = transformations

        component_i += 1
        if component_i == 3:
            # All (x,y,z) components were parsed
            # -> head to the next transformation
            transformation_i += 1
            component_i = 0

    return matrices



import bpy


# Default 5.x compositor node tree
def default_5x_compositor_node_tree():
    node_tree = bpy.data.node_groups.new(
        type="CompositorNodeTree", name="Compositor Nodes"
    )
    render_layers = node_tree.nodes.new("CompositorNodeRLayers")
    group_output = node_tree.nodes.new("NodeGroupOutput")
    _ = node_tree.interface.new_socket(
        name="Image", in_out="OUTPUT", socket_type="NodeSocketColor"
    )
    render_layers.location = (100.0, 250.0)
    group_output.location = (600.0, 250.0)
    node_tree.links.new(render_layers.outputs["Image"], group_output.inputs["Image"])
    return node_tree


# MN Compositor node tree
def mn_compositor_node_tree():
    mn_compositor = bpy.data.node_groups.new(
        type="CompositorNodeTree", name="MN Compositor"
    )

    # mn_compositor interface
    # Socket Image Output
    mn_compositor.interface.new_socket(
        name="Image", in_out="OUTPUT", socket_type="NodeSocketColor"
    )
    # Socket Image Input
    mn_compositor.interface.new_socket(
        name="Image", in_out="INPUT", socket_type="NodeSocketColor"
    )

    # initialize mn_compositor nodes
    # node Group Output
    group_output = mn_compositor.nodes.new("NodeGroupOutput")
    group_output.name = "Group Output"

    # node Group Input
    group_input = mn_compositor.nodes.new("NodeGroupInput")
    group_input.name = "Group Input"

    # node Image
    image = mn_compositor.nodes.new("CompositorNodeImage")
    image.name = "Image"
    if "mn_annotations" in bpy.data.images:
        image.image = bpy.data.images["mn_annotations"]

    # node Alpha Over
    alpha_over = mn_compositor.nodes.new("CompositorNodeAlphaOver")
    alpha_over.name = "Alpha Over"
    # Fac
    alpha_over.inputs["Fac"].default_value = 1.0

    # Set locations
    group_output.location = (380.0, 20.0)
    group_input.location = (0.0, 20.0)
    image.location = (0.0, -80.0)
    alpha_over.location = (200.0, -80.0)

    # Set dimensions
    group_output.width, group_output.height = 140.0, 100.0
    group_input.width, group_input.height = 140.0, 100.0
    image.width, image.height = 140.0, 100.0
    alpha_over.width, alpha_over.height = 140.0, 100.0

    # initialize mn_compositor links
    # alpha_over.Image -> group_output.Image
    mn_compositor.links.new(alpha_over.outputs[0], group_output.inputs[0])
    # image.Image -> alpha_over.Image
    mn_compositor.links.new(image.outputs[0], alpha_over.inputs[2])
    # group_input.Image -> alpha_over.Image
    mn_compositor.links.new(group_input.outputs[0], alpha_over.inputs[1])
    return mn_compositor



import bpy
import databpy


def get_links(socket: bpy.types.NodeSocket) -> bpy.types.NodeLinks:
    """
    Get the links between two sockets
    """
    links = socket.links
    if links is None:
        raise ValueError("No links found for socket")
    return links


def int_menu_switch_from_dict(
    dictionary: dict, name: str = "NewMenuSwitch"
) -> bpy.types.NodeTree:
    """
    Create a menu from a dictionary of options
    """
    tree = databpy.nodes.new_tree(name=name, geometry=False)

    node_m = tree.nodes.new("GeometryNodeMenuSwitch")
    node_m.data_type = "INT"

    starting_enum_length = len(node_m.enum_items)

    for i, (key, value) in enumerate(dictionary.items()):
        if i < starting_enum_length:
            node_m.enum_items[i].name = key
        else:
            node_m.enum_items.new(name=key)

        node_m.inputs[i + 1].default_value = value

    tree.interface.new_socket("Selector", socket_type="NodeSocketMenu")
    tree.interface.new_socket("Value", socket_type="NodeSocketInt", in_out="OUTPUT")
    tree.links.new(databpy.nodes.get_input(tree).outputs["Selector"], node_m.inputs[0])
    tree.links.new(
        node_m.outputs["Output"], databpy.nodes.get_output(tree).inputs["Value"]
    )
    return tree



from pathlib import Path
from . import data
from .template import install, uninstall

ASSET_DIR = Path(__file__).resolve().parent
ADDON_DIR = ASSET_DIR.parent
MN_DATA_FILE = ASSET_DIR / "node_data_file.blend"


__all__ = ["data", "install", "uninstall", "MN_DATA_FILE"]



from biotite.structure import AtomArrayStack
from ...annotations.base import BaseAnnotation
from ...annotations.manager import BaseAnnotationManager
from ..annotations import Label2D, Label3D
from ..base import EntityType


class MoleculeAnnotation(BaseAnnotation):
    """
    Base class for a Molecule Annotation

    All molecule annotations should derive from this base class and implement
    the 'draw' method. All derived classes will have access to the molecule
    instance (self.molecule) and all the annotation inputs and common params
    via self.interface.<property>

    An optional 'defaults' method can be provided to set default values
    to the annotation.

    """

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        # Auto-register any sub classes with the annotation manager
        MoleculeAnnotationManager.register(cls)

    def __init__(self, molecule):
        # Allow access to the molecule entity within the annotations
        self.molecule = molecule
        super().__init__()


class MoleculeAnnotationManager(BaseAnnotationManager):
    """
    Annotation Manager for Molecule Entity

    """

    _entity_type = EntityType.MOLECULE
    _classes = {}  # Entity class specific annotation classes

    def __init__(self, entity):
        super().__init__(entity)
        self._interfaces = {}  # Entity instance specific annotation interfaces


class MoleculeInfo(MoleculeAnnotation):
    """
    Molecule Info Annotation

    Attributes
    ----------
    location: tuple[float, float]
        Normalized coordinates (0.0 - 1.0) of the postion in viewport / render

    show_models: bool
        Whether or not to show the number of models in the molecule

    show_atoms: bool
        Whether or not to show the number of atoms in the molecule

    custom_text: str
        Any custom text to add at the end of the annotation

    """

    annotation_type = "molecule_info"

    location: tuple[float, float] = (0.025, 0.05)
    show_models: bool = True
    show_atoms: bool = True
    custom_text: str = ""

    def defaults(self) -> None:
        params = self.interface
        params.text_align = "left"

    def validate(self) -> bool:
        params = self.interface
        x, y = params.location
        if (not 0 <= x <= 1) or (not 0 <= y <= 1):
            raise ValueError("Normalized coordinates should lie between 0 and 1")
        return True

    def draw(self) -> None:
        params = self.interface
        molecule = self.molecule
        text = ""
        if params.show_models:
            text = f"Models: {molecule.n_models}"
        if params.show_atoms:
            array = molecule.array
            if isinstance(array, AtomArrayStack):
                text = text + f"|Atoms[0]: {len(array[0])}"
            else:
                text = text + f"|Atoms: {len(array)}"
        if params.custom_text != "":
            text = text + "|" + params.custom_text
        # Draw text at normalized coordinates wrt viewport / render
        self.draw_text_2d_norm(params.location, text)


class Label2D(MoleculeAnnotation, Label2D):
    """Common Label2D Annotation for all entities"""


class Label3D(MoleculeAnnotation, Label3D):
    """Common Label3D Annotation for all entities"""



from . import assets, blender, color, converters, download, nodes, session, ui
from .assets import template
from .entities import Molecule, MoleculeSelector, Trajectory
from .nodes import material
from .nodes.styles import (
    StyleBallAndStick,
    StyleCartoon,
    StyleRibbon,
    StyleSpheres,
    StyleSticks,
    StyleSurface,
)
from .ui.addon import register, unregister

try:
    from .scene import Canvas
except ModuleNotFoundError:
    pass

__all__ = [
    "assets",
    "blender",
    "color",
    "converters",
    "session",
    "ui",
    "nodes",
    "material",
    "template",
    "Molecule",
    "Trajectory",
    "register",
    "unregister",
    "download",
    "Canvas",
    "MoleculeSelector",
    "StyleRibbon",
    "StyleCartoon",
    "StyleBallAndStick",
    "StyleSpheres",
    "StyleSticks",
    "StyleSurface",
]



"""
Trajectory selection management for MolecularNodes.

This module provides classes for managing atom selections in molecular dynamics
trajectories. It bridges MDAnalysis AtomGroups with Blender's geometry node system,
allowing dynamic selections that update during trajectory playback.

Data Flow Architecture
----------------------
PropertyGroups (TrajectorySelectionItem) are the source of truth for persistent data:
    User edits UI → PropertyGroup.update callback → _update_entities() handler
                                                              ↓
                             Selection.sync_from_ui() ← Called by handler
                                                              ↓
                             Selection.set_atom_group() ← Recomputes _ag
                                                              ↓
                             Selection.sync_to_blender_attribute() ← Updates geometry

Classes
-------
SelectionError
    Exception raised when selection operations fail.
Selection
    Wraps an MDAnalysis AtomGroup and syncs it with Blender UI properties
    and geometry node attributes.
SelectionManager
    Manages all selections for a trajectory, coordinating between Python objects,
    Blender UI, and geometry nodes.

Notes
-----
Selections can be either static (computed once) or updating (recalculated every frame).
The selection state is stored as boolean mask arrays that are exposed as geometry
node attributes for use in Blender's node-based rendering system.
"""

from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ...ui.props import TrajectorySelectionItem
    from .base import Trajectory
from typing import Dict
from uuid import uuid1
import bpy
import databpy as db
import MDAnalysis as mda
import numpy as np
import numpy.typing as npt


class SelectionError(Exception):
    """Exception raised when selection operations fail."""

    pass


class AtomGroupMetadata:
    """Metadata extracted from an MDAnalysis AtomGroup."""

    def __init__(self, atomgroup: mda.AtomGroup):
        """
        Initialize metadata from an AtomGroup.

        Parameters
        ----------
        atomgroup : mda.AtomGroup
            The AtomGroup to extract metadata from.
        """
        self._atomgroup = atomgroup
        self._uuid = str(uuid1())

    @property
    def fallback_name(self) -> str:
        return self.string or f"sel_{self._atomgroup.n_atoms}_atoms"

    @property
    def uuid(self) -> str:
        """Get the unique identifier for this selection."""
        return self._uuid

    @property
    def string(self) -> str:
        """Extract the selection string from the AtomGroup."""
        # Assuming it's a single selection
        # MDA does support `u.select_atoms('index 0', 'around 5 index 0')`
        try:
            return self._atomgroup._selection_strings[0]
        except (AttributeError, IndexError):
            return ""

    @property
    def updating(self) -> bool:
        """Determine if the AtomGroup is updating."""
        return self._atomgroup.__class__.__name__ == "UpdatingAtomGroup"

    @property
    def periodic(self) -> bool:
        """Determine if periodic boundary conditions are enabled."""
        if not self.updating:
            return False
        try:
            return self._atomgroup._selections[0].periodic
        except (AttributeError, IndexError):
            # Some selections don't have the periodic attribute
            return False
        except Exception as e:
            print(f"Warning extracting periodic setting: {e}")
            return False


class Selection:
    """
    Represents an atom selection for a trajectory in Blender.

    A Selection wraps an MDAnalysis AtomGroup and provides synchronization between
    the MDAnalysis selection and Blender UI properties. It manages the selection string,
    updating behavior, and conversion to boolean masks for geometry node attributes.

    Parameters
    ----------
    trajectory : Trajectory
        The parent Trajectory object this selection belongs to.
    name : str, optional
        Unique name for this selection. Default is "selection_0".

    Attributes
    ----------
    trajectory : Trajectory
        Reference to the parent trajectory.
    mask_array : npt.NDArray[np.bool_]
        Cached boolean mask representing which atoms are selected.
    """

    def __init__(self, trajectory: Trajectory, name: str = "selection_0"):
        self._ag: mda.AtomGroup | None = None
        self._uuid: str = str(uuid1())
        self._name = name
        self._current_string: str = ""
        self.trajectory = trajectory
        self._ui_item_cache: TrajectorySelectionItem | None = None

    def add_selection_property(self, string: str = "all", updating=True, periodic=True):
        """
        Initialize the selection with a UI property and create the atom group.

        Creates a Blender UI property item for this selection and sets up the
        MDAnalysis AtomGroup based on the provided selection string.

        Parameters
        ----------
        string : str, optional
            MDAnalysis selection string. Default is "all".
        updating : bool, optional
            Whether the selection should update dynamically during trajectory playback.
            Default is True.
        periodic : bool, optional
            Whether to consider periodic boundary conditions in the selection.
            Default is True.
        """
        prop = self.trajectory.selections.items.add()
        prop.name = self.name
        prop.uuid = self._uuid
        self.updating = updating
        self.periodic = periodic
        self.set_atom_group(string)
        self.string = string
        self.mask_array = self._ag_to_mask()

    @property
    def name(self) -> str:
        """
        Get the name of this selection.

        Returns
        -------
        str
            The unique name identifying this selection.
        """
        return self._name

    @property
    def ui_item(self) -> "TrajectorySelectionItem":
        """
        Get the Blender UI property item for this selection.

        Uses caching to avoid repeated lookups. The cache can be invalidated
        by calling invalidate_ui_cache().

        Returns
        -------
        TrajectorySelectionItem
            The Blender property group containing UI-accessible properties
            for this selection (string, updating, periodic, message, etc.).

        Raises
        ------
        RuntimeError
            If the UI property cannot be found, indicating the selection
            was not properly initialized.
        """
        if self._ui_item_cache is None:
            try:
                self._ui_item_cache = self.trajectory.selections.items[self.name]
            except (KeyError, AttributeError) as e:
                raise RuntimeError(
                    f"UI property not found for selection '{self.name}'. "
                    f"Selection may not be properly initialized."
                ) from e
        return self._ui_item_cache

    def invalidate_ui_cache(self) -> None:
        """
        Invalidate the cached UI property reference.

        Call this if the UI item is recreated or the selection is renamed.
        The next access to ui_item will fetch a fresh reference.
        """
        self._ui_item_cache = None

    @property
    def string(self) -> str:
        """
        Get the MDAnalysis selection string.

        Returns
        -------
        str
            The current selection string (e.g., "name CA", "resid 1:10").
        """
        return self.ui_item.string

    @string.setter
    def string(self, string: str) -> None:
        """
        Set the MDAnalysis selection string and update the atom group.

        Updates the selection string in the UI property and attempts to
        create a new AtomGroup. If successful, updates the selection mask
        in Blender. If an error occurs, stores the error message.

        Parameters
        ----------
        string : str
            MDAnalysis selection string to apply.

        Notes
        -----
        If the selection is marked as immutable, logs a warning and returns
        without making changes.
        """
        if self.immutable:
            print(
                f"Warning: Attempted to modify immutable selection '{self.name}'. "
                "Ignoring change."
            )
            return

        self.ui_item.string = string
        success = self.apply_selection_string(string)
        if success:
            self.sync_to_blender_attribute()

    @property
    def periodic(self) -> bool:
        """
        Get whether periodic boundary conditions are considered.

        Returns
        -------
        bool
            True if periodic boundary conditions are enabled for this selection.
        """
        return self.ui_item.periodic

    @periodic.setter
    def periodic(self, periodic: bool) -> None:
        """
        Set whether to consider periodic boundary conditions.

        Parameters
        ----------
        periodic : bool
            If True, periodic boundary conditions will be considered in the selection.
        """
        if self.ui_item.periodic == periodic:
            return
        self.ui_item.periodic = periodic

    @property
    def updating(self) -> bool:
        """
        Get whether this selection updates dynamically during playback.

        Returns
        -------
        bool
            True if the selection recalculates every frame during trajectory playback.
        """
        return self.ui_item.updating

    @updating.setter
    def updating(self, updating: bool) -> None:
        """
        Set whether this selection should update dynamically during playback.

        Parameters
        ----------
        updating : bool
            If True, the selection will recalculate every frame. If False,
            the selection is static and uses the mask from the initial frame.
        """
        if self.ui_item.updating == updating:
            return
        self.ui_item.updating = updating

    @property
    def message(self) -> str:
        """
        Get the current status or error message for this selection.

        Returns
        -------
        str
            Error message if selection parsing failed, empty string otherwise.
        """
        return self.ui_item.message

    @message.setter
    def message(self, message: str) -> None:
        """
        Set the status or error message for this selection.

        Parameters
        ----------
        message : str
            Message to display to the user, typically an error from selection parsing.
        """
        self.ui_item.message = message

    @property
    def immutable(self) -> bool:
        """
        Get whether this selection can be modified by the user.

        Returns
        -------
        bool
            True if the selection is locked from user modification.
        """
        return self.ui_item.immutable

    @immutable.setter
    def immutable(self, immutable: bool) -> None:
        """
        Set whether this selection can be modified by the user.

        Parameters
        ----------
        immutable : bool
            If True, prevents the user from modifying this selection in the UI.
        """
        self.ui_item.immutable = immutable

    def validate_selection_string(self, string: str) -> tuple[bool, str]:
        """
        Validate a selection string without side effects.

        Tests whether the selection string can be parsed by MDAnalysis
        without actually storing the result.

        Parameters
        ----------
        string : str
            MDAnalysis selection string to validate.

        Returns
        -------
        tuple[bool, str]
            A tuple of (is_valid, error_message). If valid, error_message is empty.

        Examples
        --------
        >>> is_valid, error = sel.validate_selection_string("name CA")
        >>> if is_valid:
        ...     sel.apply_selection_string("name CA")
        """
        try:
            # Just test parsing, don't store
            self.trajectory.universe.select_atoms(
                string, updating=False, periodic=False
            )
            return True, ""
        except Exception as e:
            return False, str(e)

    def apply_selection_string(self, string: str) -> bool:
        """
        Apply a selection string and update the atom group.

        This method validates the selection string and, if valid, creates
        the AtomGroup and updates the cached mask array.

        Parameters
        ----------
        string : str
            MDAnalysis selection string to apply.

        Returns
        -------
        bool
            True if the selection was successfully applied, False otherwise.

        Notes
        -----
        Unlike set_atom_group, this method returns a success status and
        handles errors gracefully by storing them in the message property.
        """
        if self._current_string == string:
            return True

        try:
            self._ag = self.trajectory.universe.select_atoms(
                string, updating=self.updating, periodic=self.periodic
            )
            self.mask_array = self._ag_to_mask()
            self._current_string = string
            self.message = ""
            return True
        except Exception as e:
            error_msg = (
                f"{str(e)} in selection: `{self.name}` "
                f"on object: `{self.trajectory.object.name}`"
            )
            self.message = str(e)
            print(error_msg)
            return False

    def set_atom_group(self, string: str, raise_on_error: bool = False) -> bool:
        """
        Create or update the MDAnalysis AtomGroup from a selection string.

        Parses the selection string using MDAnalysis and creates an AtomGroup
        (static or updating depending on the updating property). Updates the
        cached mask array if successful.

        Parameters
        ----------
        string : str
            MDAnalysis selection string to parse.
        raise_on_error : bool, optional
            If True, raises SelectionError on failure.
            If False, stores error in message and returns False.
            Default is False.

        Returns
        -------
        bool
            True if successful, False if failed (only when raise_on_error=False).

        Raises
        ------
        SelectionError
            If raise_on_error=True and the selection string is invalid.

        Notes
        -----
        This method is maintained for backward compatibility. New code should
        prefer apply_selection_string() for better error handling.
        """
        if self._current_string == string:
            return True

        try:
            self._ag = self.trajectory.universe.select_atoms(
                string, updating=self.updating, periodic=self.periodic
            )
            self.mask_array = self._ag_to_mask()
            self._current_string = string
            self.message = ""
            return True
        except Exception as e:
            error_msg = (
                f"{str(e)} in selection: `{self.name}` "
                f"on object: `{self.trajectory.object.name}`"
            )
            self.message = str(e)
            print(error_msg)

            if raise_on_error:
                raise SelectionError(error_msg) from e
            return False

    def _ag_to_mask(self) -> npt.NDArray[np.bool_]:
        """
        Convert the AtomGroup to a boolean mask array.

        Creates a boolean mask where True indicates atoms in the selection and
        False indicates atoms not in the selection. The mask is aligned with the
        trajectory's Universe atom ordering.

        Returns
        -------
        npt.NDArray[np.bool_]
            1D boolean array with length equal to the number of atoms in the
            Universe, where True indicates the atom is in this selection.
        """
        return np.isin(self.trajectory.universe.atoms.ix, self._ag.ix).astype(bool)

    def sync_to_blender_attribute(self) -> None:
        """
        Synchronize the selection mask to Blender geometry attributes.

        Converts the current selection to a boolean mask and stores it as a
        geometry node attribute on the trajectory object. Only applies if
        updating=True.

        Notes
        -----
        This pushes the Python-side selection state to Blender's geometry system,
        making it available for use in geometry nodes.
        """
        if not self.updating:
            return
        mask = self.to_mask()
        self.trajectory.store_named_attribute(
            mask, name=self.name, atype=db.AttributeTypes.BOOLEAN
        )

    def set_selection(self) -> None:
        """
        Store the selection mask as a named attribute in Blender.

        .. deprecated::
            Use sync_to_blender_attribute() instead for clearer intent.

        Notes
        -----
        This method is maintained for backward compatibility.
        """
        self.sync_to_blender_attribute()

    def sync_from_ui(self) -> None:
        """
        Synchronize selection state from the UI property to computed state.

        Pulls the current state from the Blender UI property and updates
        the computed AtomGroup if the selection string has changed.

        Notes
        -----
        This is called by update handlers when the user modifies the selection
        through Blender's UI. It ensures the Python-side state reflects
        the PropertyGroup state.
        """
        self.invalidate_ui_cache()
        # Recompute AtomGroup if string changed
        if self.ui_item.string != self._current_string:
            self.apply_selection_string(self.ui_item.string)
            self.sync_to_blender_attribute()

    def to_mask(self) -> npt.NDArray[np.bool_]:
        """
        Get the selection as a boolean mask array.

        Returns the cached mask array. If updating=True, recomputes the mask
        from the current frame's AtomGroup state before returning.

        Returns
        -------
        npt.NDArray[np.bool_]
            1D boolean array indicating which atoms are selected.
        """
        if self.updating:
            self.mask_array = self._ag_to_mask()
        return self.mask_array

    def is_valid(self) -> bool:
        """
        Check if the selection is in a valid state.

        Verifies that all required components are present: the AtomGroup,
        the mask array, and the UI property.

        Returns
        -------
        bool
            True if the selection is valid and usable, False otherwise.

        Examples
        --------
        >>> if not sel.is_valid():
        ...     print("Selection needs reinitialization")
        """
        try:
            has_ag = self._ag is not None
            has_mask = self.mask_array is not None
            has_ui = self.name in self.trajectory.selections.items
            return has_ag and has_mask and has_ui
        except Exception:
            return False

    def ensure_valid(self) -> None:
        """
        Raise an exception if the selection is in an invalid state.

        Useful for catching bugs early in development or providing
        clear error messages when a selection becomes corrupted.

        Raises
        ------
        SelectionError
            If the selection is missing required components or is otherwise invalid.

        Examples
        --------
        >>> sel.ensure_valid()  # Raises if invalid
        >>> # Continue with valid selection operations
        """
        if not self.is_valid():
            has_ag = self._ag is not None
            has_mask = self.mask_array is not None
            try:
                has_ui = self.name in self.trajectory.selections.items
            except Exception:
                has_ui = False

            raise SelectionError(
                f"Selection '{self.name}' is in an invalid state. "
                f"Has AtomGroup: {has_ag}, "
                f"Has mask: {has_mask}, "
                f"In UI: {has_ui}"
            )

    @classmethod
    def from_atomgroup(cls, trajectory, atomgroup: mda.AtomGroup, name: str = ""):
        """
        Create a Selection from an existing MDAnalysis AtomGroup.

        This factory method creates a Selection object from an existing AtomGroup,
        automatically detecting whether it's an UpdatingAtomGroup and extracting
        the selection string and periodic settings if available.

        Parameters
        ----------
        trajectory : Trajectory
            The parent trajectory this selection belongs to.
        atomgroup : mda.AtomGroup
            An MDAnalysis AtomGroup or UpdatingAtomGroup to wrap.
        name : str, optional
            Name for the selection. If empty, uses the selection string or
            a default name based on atom count.

        Returns
        -------
        Selection
            A new Selection object wrapping the provided AtomGroup.

        Notes
        -----
        - If atomgroup is an UpdatingAtomGroup, extracts the selection string
          and periodic settings from it.
        - The selection is marked as immutable to prevent user modification.
        - The selection is automatically registered with the trajectory's
          SelectionManager.
        - Creates the Python wrapper FIRST and registers it BEFORE creating the UI
          property, to avoid race conditions with update handlers.
        """
        # Extract metadata from the AtomGroup using the class directly
        meta = AtomGroupMetadata(atomgroup)
        name = meta.fallback_name if name == "" else name

        # Create Python wrapper FIRST
        sel = cls(trajectory=trajectory, name=name)
        sel._uuid = meta.uuid
        sel._ag = atomgroup
        sel.mask_array = sel._ag_to_mask()
        sel._current_string = meta.string

        # Register with manager BEFORE creating UI property
        # This prevents race conditions when the UI property setter triggers update handlers
        trajectory.selections.append(sel)

        # Create UI property LAST
        # Note: Setting prop.name may trigger update handlers, but the selection
        # is already registered in _selections, so get() will find it
        prop = trajectory.selections.items.add()
        prop.name = name
        prop.uuid = meta.uuid
        prop.string = meta.string
        prop.updating = meta.updating
        prop.periodic = meta.periodic
        prop.immutable = True

        return sel


class SelectionManager:
    """
    Manages all atom selections for a trajectory.

    The SelectionManager coordinates between Selection objects (Python side),
    Blender UI properties, and geometry node attributes. It provides methods
    to create, retrieve, and remove selections, and maintains synchronization
    between the different representation layers.

    Parameters
    ----------
    trajectory : Trajectory
        The parent trajectory object this manager belongs to.

    Attributes
    ----------
    trajectory : Trajectory
        Reference to the parent trajectory.
    """

    def __init__(self, trajectory: "Trajectory"):
        self.trajectory = trajectory
        self._selections: Dict[str, Selection] = {}

    def _validate_name_unique(self, name: str) -> None:
        """
        Validate that a selection name is unique.

        Parameters
        ----------
        name : str
            The name to validate.

        Raises
        ------
        ValueError
            If a selection with this name already exists.
        """
        if name in self._selections:
            raise ValueError(
                f"Selection '{name}' already exists. "
                "Use a unique name or remove the existing selection first."
            )

    def add(
        self,
        string: str,
        name: str = "selection_0",
        updating: bool = True,
        periodic: bool = False,
    ) -> Selection:
        """
        Create and register a new selection from a selection string.

        Creates a new Selection object with the specified parameters and
        registers it with both the manager's internal dictionary and the
        Blender UI property collection.

        Parameters
        ----------
        string : str
            MDAnalysis selection string (e.g., "name CA", "resid 1:10").
        name : str, optional
            Unique name for this selection. Default is "selection_0".
        updating : bool, optional
            Whether the selection updates every frame. Default is True.
        periodic : bool, optional
            Whether to consider periodic boundary conditions. Default is False.

        Returns
        -------
        Selection
            The newly created and registered Selection object.

        Raises
        ------
        ValueError
            If a selection with the given name already exists.
        """
        # Note: Not validating uniqueness here for backward compatibility
        # as existing code may rely on overwriting selections
        sel = self._selections[name] = Selection(self.trajectory, name=name)
        sel.add_selection_property(string=string, updating=updating, periodic=periodic)
        return sel

    def create_selection(
        self,
        string: str,
        name: str = "selection_0",
        updating: bool = True,
        periodic: bool = False,
        validate_unique: bool = True,
    ) -> Selection:
        """
        Create and register a new selection with validation.

        This is the preferred method for creating new selections as it
        includes proper validation.

        Parameters
        ----------
        string : str
            MDAnalysis selection string (e.g., "name CA", "resid 1:10").
        name : str, optional
            Unique name for this selection. Default is "selection_0".
        updating : bool, optional
            Whether the selection updates every frame. Default is True.
        periodic : bool, optional
            Whether to consider periodic boundary conditions. Default is False.
        validate_unique : bool, optional
            Whether to validate that the name is unique. Default is True.

        Returns
        -------
        Selection
            The newly created and registered Selection object.

        Raises
        ------
        ValueError
            If validate_unique=True and a selection with the given name already exists.
        """
        if validate_unique:
            self._validate_name_unique(name)

        return self.add(string=string, name=name, updating=updating, periodic=periodic)

    def register_selection(self, selection: Selection) -> None:
        """
        Register an existing Selection object with this manager.

        Adds a Selection object to the internal dictionary without creating
        a new UI property (assumes the Selection was created via a classmethod
        that already handled UI property creation).

        Parameters
        ----------
        selection : Selection
            The Selection object to register.

        Notes
        -----
        This method is typically used internally by factory methods like
        from_atomgroup() after they've created both the UI property and
        the Selection object.
        """
        self._selections[selection.name] = selection

    def append(self, selection: Selection) -> None:
        """
        Register an existing Selection object with this manager.

        .. deprecated::
            Use register_selection() instead for clearer intent.

        Parameters
        ----------
        selection : Selection
            The Selection object to register.
        """
        self.register_selection(selection)

    def from_ui_item(self, item: TrajectorySelectionItem) -> Selection:
        """
        Create a Selection from a Blender UI property item.

        Creates a new Selection based on the properties stored in a Blender
        UI property item, typically used when loading from a saved .blend file.

        Parameters
        ----------
        item : TrajectorySelectionItem
            Blender property group containing selection parameters.

        Returns
        -------
        Selection
            The newly created Selection object.
        """
        return self.add(
            item.string, item.name, updating=item.updating, periodic=item.periodic
        )

    def from_atomgroup(
        self, atomgroup: mda.AtomGroup, name: str = "NewSelection"
    ) -> Selection:
        """
        Create a Selection from an existing MDAnalysis AtomGroup.

        Wraps an AtomGroup in a Selection object and stores the initial mask
        as a Blender geometry attribute. Useful for creating selections from
        programmatically defined AtomGroups.

        Parameters
        ----------
        atomgroup : mda.AtomGroup
            MDAnalysis AtomGroup or UpdatingAtomGroup to wrap.
        name : str, optional
            Name for the new selection. Default is "NewSelection".

        Returns
        -------
        Selection
            The newly created Selection object.

        See Also
        --------
        Selection.from_atomgroup : The underlying classmethod used to create the Selection.
        """
        sel = Selection.from_atomgroup(
            trajectory=self.trajectory, atomgroup=atomgroup, name=name
        )
        # Note: sel is already registered in _selections by Selection.from_atomgroup()
        # Store the mask as a Blender attribute
        self.trajectory.store_named_attribute(
            sel.to_mask(), name=sel.name, atype=db.AttributeTypes.BOOLEAN
        )
        return sel

    def remove(self, name: str) -> None:
        """
        Remove a selection from the manager and clean up all references.

        Removes the selection from the internal dictionary, the Blender UI
        property collection, and the geometry node attribute. Handles cases
        where the selection may be partially registered.

        Parameters
        ----------
        name : str
            Name of the selection to remove.

        Notes
        -----
        Silently ignores KeyError or AttributeError if the selection was not
        fully registered in all locations.
        """
        names = [sel.name for sel in self.items]
        index = names.index(name)
        self.items.remove(index)
        try:
            del self._selections[name]
        except KeyError:
            pass
        try:
            self.trajectory.remove_named_attribute(name)
        except AttributeError:
            pass

    def get(self, name: str, lazy_init: bool = True) -> Selection | None:
        """
        Retrieve a Selection by name.

        If the selection exists in the UI properties but not in the Python
        runtime (common when loading from saved .blend files), it will be
        lazily initialized from the UI property.

        Parameters
        ----------
        name : str
            Name of the selection to retrieve.
        lazy_init : bool, optional
            If True, attempts to initialize the selection from UI properties
            if it doesn't exist in Python runtime. Default is True.

        Returns
        -------
        Selection
            The Selection object with the given name.

        Raises
        ------
        KeyError
            If no selection with the given name exists in either Python
            runtime or UI properties.

        Notes
        -----
        Lazy initialization ensures that selections persist correctly when
        loading .blend files, where UI properties are restored but Python
        objects need to be reconstructed.
        """
        # Fast path: already in runtime
        if name in self._selections:
            return self._selections.get(name)

        # Lazy initialization path
        if lazy_init:
            try:
                # Check if UI property exists
                ui_item = self.items[name]
                # Initialize from UI property
                selection = self.from_ui_item(ui_item)
                return selection
            except (KeyError, AttributeError):
                # UI property doesn't exist either
                pass

        # Selection not found anywhere
        raise KeyError(
            f"Selection '{name}' not found. "
            f"Available selections: {list(self._selections.keys())}"
        )

    @property
    def items(self) -> bpy.types.CollectionProperty:
        """
        Get the Blender UI property collection for all selections.

        Returns
        -------
        bpy.types.CollectionProperty
            The Blender property collection containing TrajectorySelectionItem
            properties for this trajectory.
        """
        return self.trajectory.object.mn_trajectory_selections

    @property
    def index(self) -> int:
        """
        Get the currently selected index in the UI list.

        Returns
        -------
        int
            The index of the currently selected selection in the UI list.
        """
        return self.trajectory.object.mn["list_index"]

    @index.setter
    def index(self, value: int) -> None:
        """
        Set the currently selected index in the UI list.

        Parameters
        ----------
        value : int
            The index to select in the UI list.
        """
        self.trajectory.object.mn["list_index"] = value

    def __len__(self) -> int:
        """
        Get the number of selections managed by this manager.

        Returns
        -------
        int
            The number of registered Selection objects.
        """
        return len(self._selections)

    def sync_all_from_ui(self) -> None:
        """
        Synchronize all selections from UI properties.

        Iterates through all UI property items and ensures that each has
        a corresponding Python Selection object. This is useful when loading
        from saved .blend files where the UI properties are restored but
        Python objects need to be reconstructed.

        Notes
        -----
        This method is typically called during trajectory initialization
        or when loading from disk. It ensures that the Python runtime state
        matches the persistent UI property state.
        """
        for ui_item in self.items:
            if ui_item.name not in self._selections:
                try:
                    self.from_ui_item(ui_item)
                except Exception as e:
                    print(
                        f"Warning: Failed to initialize selection '{ui_item.name}': {e}"
                    )

    def has_selection(self, name: str) -> bool:
        """
        Check if a selection exists.

        Checks both the Python runtime and UI properties for the selection.

        Parameters
        ----------
        name : str
            Name of the selection to check.

        Returns
        -------
        bool
            True if the selection exists, False otherwise.
        """
        if name in self._selections:
            return True
        try:
            return name in self.items
        except Exception:
            return False
